# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import strutils
import webob

from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.api.schemas import quotas
from cinder.api import validation
from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder.policies import quotas as policy
from cinder import quota
from cinder import quota_utils
from cinder import utils

LOG = logging.getLogger(__name__)

QUOTAS = quota.QUOTAS
GROUP_QUOTAS = quota.GROUP_QUOTAS
NON_QUOTA_KEYS = quota.NON_QUOTA_KEYS


class QuotaSetsController(wsgi.Controller):

    def _format_quota_set(self, project_id, quota_set):
        """Convert the quota object to a result dict."""

        quota_set['id'] = str(project_id)

        return dict(quota_set=quota_set)

    def _validate_existing_resource(self, key, value, quota_values):
        # -1 limit will always be greater than the existing value
        if key == 'per_volume_gigabytes' or value == -1:
            return
        v = quota_values.get(key, {})
        used = (v.get('in_use', 0) + v.get('reserved', 0))
        if QUOTAS.using_nested_quotas():
            used += v.get('allocated', 0)
        if value < used:
            msg = (_("Quota %(key)s limit must be equal or greater than "
                     "existing resources. Current usage is %(usage)s "
                     "and the requested limit is %(limit)s.")
                   % {'key': key,
                      'usage': used,
                      'limit': value})
            raise webob.exc.HTTPBadRequest(explanation=msg)

    def _get_quotas(self, context, id, usages=False):
        values = QUOTAS.get_project_quotas(context, id, usages=usages)
        group_values = GROUP_QUOTAS.get_project_quotas(context, id,
                                                       usages=usages)
        values.update(group_values)

        if usages:
            return values
        else:
            return {k: v['limit'] for k, v in values.items()}

    def _authorize_update_or_delete(self, context_project,
                                    target_project_id,
                                    parent_id):
        """Checks if update or delete are allowed in the current hierarchy.

        With hierarchical projects, only the admin of the parent or the root
        project has privilege to perform quota update and delete operations.

        :param context_project: The project in which the user is scoped to.
        :param target_project_id: The id of the project in which the
                                  user want to perform an update or
                                  delete operation.
        :param parent_id: The parent id of the project in which the user
                          want to perform an update or delete operation.
        """
        if context_project.is_admin_project:
            # The calling project has admin privileges and should be able
            # to operate on all quotas.
            return
        if context_project.parent_id and parent_id != context_project.id:
            msg = _("Update and delete quota operations can only be made "
                    "by an admin of immediate parent or by the CLOUD admin.")
            raise webob.exc.HTTPForbidden(explanation=msg)

        if context_project.id != target_project_id:
            if not self._is_descendant(target_project_id,
                                       context_project.subtree):
                msg = _("Update and delete quota operations can only be made "
                        "to projects in the same hierarchy of the project in "
                        "which users are scoped to.")
                raise webob.exc.HTTPForbidden(explanation=msg)
        else:
            msg = _("Update and delete quota operations can only be made "
                    "by an admin of immediate parent or by the CLOUD admin.")
            raise webob.exc.HTTPForbidden(explanation=msg)

    def _authorize_show(self, context_project, target_project):
        """Checks if show is allowed in the current hierarchy.

        With hierarchical projects, users are allowed to perform a quota show
        operation if they have the cloud admin role or if they belong to at
        least one of the following projects: the target project, its immediate
        parent project, or the root project of its hierarchy.

        :param context_project: The project in which the user
                                is scoped to.
        :param target_project: The project in which the user wants
                               to perform a show operation.
        """
        if context_project.is_admin_project:
            # The calling project has admin privileges and should be able
            # to view all quotas.
            return
        if target_project.parent_id:
            if target_project.id != context_project.id:
                if not self._is_descendant(target_project.id,
                                           context_project.subtree):
                    msg = _("Show operations can only be made to projects in "
                            "the same hierarchy of the project in which users "
                            "are scoped to.")
                    raise webob.exc.HTTPForbidden(explanation=msg)
                if context_project.id != target_project.parent_id:
                    if context_project.parent_id:
                        msg = _("Only users with token scoped to immediate "
                                "parents or root projects are allowed to see "
                                "its children quotas.")
                        raise webob.exc.HTTPForbidden(explanation=msg)
        elif context_project.parent_id:
            msg = _("An user with a token scoped to a subproject is not "
                    "allowed to see the quota of its parents.")
            raise webob.exc.HTTPForbidden(explanation=msg)

    def _is_descendant(self, target_project_id, subtree):
        if subtree is not None:
            for key, value in subtree.items():
                if key == target_project_id:
                    return True
                if self._is_descendant(target_project_id, value):
                    return True
        return False

    def show(self, req, id):
        """Show quota for a particular tenant

        This works for hierarchical and non-hierarchical projects. For
        hierarchical projects admin of current project, immediate
        parent of the project or the CLOUD admin are able to perform
        a show.

        :param req: request
        :param id: target project id that needs to be shown
        """
        context = req.environ['cinder.context']
        params = req.params
        target_project_id = id
        context.authorize(policy.SHOW_POLICY,
                          target={'project_id': target_project_id})

        if not hasattr(params, '__call__') and 'usage' in params:
            usage = utils.get_bool_param('usage', params)
        else:
            usage = False

        if QUOTAS.using_nested_quotas():
            # With hierarchical projects, only the admin of the current project
            # or the root project has privilege to perform quota show
            # operations.
            target_project = quota_utils.get_project_hierarchy(
                context, target_project_id)
            context_project = quota_utils.get_project_hierarchy(
                context, context.project_id, subtree_as_ids=True,
                is_admin_project=context.is_admin)

            self._authorize_show(context_project, target_project)

        quotas = self._get_quotas(context, target_project_id, usage)
        return self._format_quota_set(target_project_id, quotas)

    @validation.schema(quotas.update_quota)
    def update(self, req, id, body):
        """Update Quota for a particular tenant

        This works for hierarchical and non-hierarchical projects. For
        hierarchical projects only immediate parent admin or the
        CLOUD admin are able to perform an update.

        :param req: request
        :param id: target project id that needs to be updated
        :param body: key, value pair that will be applied to
                     the resources if the update succeeds
        """
        context = req.environ['cinder.context']
        target_project_id = id
        context.authorize(policy.UPDATE_POLICY,
                          target={'project_id': target_project_id})
        self.validate_string_length(id, 'quota_set_name',
                                    min_length=1, max_length=255)

        # Saving off this value since we need to use it multiple times
        use_nested_quotas = QUOTAS.using_nested_quotas()
        if use_nested_quotas:
            # Get the parent_id of the target project to verify whether we are
            # dealing with hierarchical namespace or non-hierarchical namespace
            target_project = quota_utils.get_project_hierarchy(
                context, target_project_id, parents_as_ids=True)
            parent_id = target_project.parent_id

            if parent_id:
                # Get the children of the project which the token is scoped to
                # in order to know if the target_project is in its hierarchy.
                context_project = quota_utils.get_project_hierarchy(
                    context, context.project_id, subtree_as_ids=True,
                    is_admin_project=context.is_admin)
                self._authorize_update_or_delete(context_project,
                                                 target_project.id,
                                                 parent_id)

        # NOTE(ankit): Pass #1 - In this loop for body['quota_set'].keys(),
        # we validate the quota limits to ensure that we can bail out if
        # any of the items in the set is bad. Meanwhile we validate value
        # to ensure that the value can't be lower than number of existing
        # resources.
        quota_values = QUOTAS.get_project_quotas(context, target_project_id,
                                                 defaults=False)
        group_quota_values = GROUP_QUOTAS.get_project_quotas(context,
                                                             target_project_id,
                                                             defaults=False)
        quota_values.update(group_quota_values)
        valid_quotas = {}
        reservations = []
        for key in body['quota_set'].keys():
            if key in NON_QUOTA_KEYS:
                continue
            self._validate_existing_resource(key, body['quota_set'][key],
                                             quota_values)

            if use_nested_quotas:
                try:
                    reservations += self._update_nested_quota_allocated(
                        context, target_project, quota_values, key,
                        body['quota_set'][key])
                except exception.OverQuota as e:
                    if reservations:
                        db.reservation_rollback(context, reservations)
                    raise webob.exc.HTTPBadRequest(explanation=e.msg)

            valid_quotas[key] = body['quota_set'][key]

        # NOTE(ankit): Pass #2 - At this point we know that all the keys and
        # values are valid and we can iterate and update them all in one shot
        # without having to worry about rolling back etc as we have done
        # the validation up front in the 2 loops above.
        for key, value in valid_quotas.items():
            try:
                db.quota_update(context, target_project_id, key, value)
            except exception.ProjectQuotaNotFound:
                db.quota_create(context, target_project_id, key, value)
            except exception.AdminRequired:
                raise webob.exc.HTTPForbidden()

        if reservations:
            db.reservation_commit(context, reservations)
        return {'quota_set': self._get_quotas(context, target_project_id)}

    def _get_quota_usage(self, quota_obj):
        return (quota_obj.get('in_use', 0) + quota_obj.get('allocated', 0) +
                quota_obj.get('reserved', 0))

    def _update_nested_quota_allocated(self, ctxt, target_project,
                                       target_project_quotas, res, new_limit):
        reservations = []
        # per_volume_gigabytes doesn't make sense to nest
        if res == "per_volume_gigabytes":
            return reservations

        quota_for_res = target_project_quotas.get(res, {})
        orig_quota_from_target_proj = quota_for_res.get('limit', 0)
        # If limit was -1, we were "taking" current child's usage from parent
        if orig_quota_from_target_proj == -1:
            orig_quota_from_target_proj = self._get_quota_usage(quota_for_res)

        new_quota_from_target_proj = new_limit
        # If we set limit to -1, we will "take" the current usage from parent
        if new_limit == -1:
            new_quota_from_target_proj = self._get_quota_usage(quota_for_res)

        res_change = new_quota_from_target_proj - orig_quota_from_target_proj
        if res_change != 0:
            deltas = {res: res_change}
            resources = QUOTAS.resources
            resources.update(GROUP_QUOTAS.resources)
            reservations += quota_utils.update_alloc_to_next_hard_limit(
                ctxt, resources, deltas, res, None, target_project.id)

        return reservations

    def defaults(self, req, id):
        context = req.environ['cinder.context']
        context.authorize(policy.SHOW_POLICY, target={'project_id': id})
        defaults = QUOTAS.get_defaults(context, project_id=id)
        group_defaults = GROUP_QUOTAS.get_defaults(context, project_id=id)
        defaults.update(group_defaults)
        return self._format_quota_set(id, defaults)

    def delete(self, req, id):
        """Delete Quota for a particular tenant.

        This works for hierarchical and non-hierarchical projects. For
        hierarchical projects only immediate parent admin or the
        CLOUD admin are able to perform a delete.

        :param req: request
        :param id: target project id that needs to be deleted
        """
        context = req.environ['cinder.context']
        context.authorize(policy.DELETE_POLICY, target={'project_id': id})

        if QUOTAS.using_nested_quotas():
            self._delete_nested_quota(context, id)
        else:
            db.quota_destroy_by_project(context, id)

    def _delete_nested_quota(self, ctxt, proj_id):
        # Get the parent_id of the target project to verify whether we are
        # dealing with hierarchical namespace or non-hierarchical
        # namespace.
        try:
            project_quotas = QUOTAS.get_project_quotas(
                ctxt, proj_id, usages=True, defaults=False)
            project_group_quotas = GROUP_QUOTAS.get_project_quotas(
                ctxt, proj_id, usages=True, defaults=False)
            project_quotas.update(project_group_quotas)
        except exception.NotAuthorized:
            raise webob.exc.HTTPForbidden()

        target_project = quota_utils.get_project_hierarchy(
            ctxt, proj_id)
        parent_id = target_project.parent_id
        if parent_id:
            # Get the children of the project which the token is scoped to
            # in order to know if the target_project is in its hierarchy.
            context_project = quota_utils.get_project_hierarchy(
                ctxt, ctxt.project_id, subtree_as_ids=True)
            self._authorize_update_or_delete(context_project,
                                             target_project.id,
                                             parent_id)

        defaults = QUOTAS.get_defaults(ctxt, proj_id)
        defaults.update(GROUP_QUOTAS.get_defaults(ctxt, proj_id))
        # If the project which is being deleted has allocated part of its
        # quota to its subprojects, then subprojects' quotas should be
        # deleted first.
        for res, value in project_quotas.items():
            if 'allocated' in project_quotas[res].keys():
                if project_quotas[res]['allocated'] > 0:
                    msg = _("About to delete child projects having "
                            "non-zero quota. This should not be performed")
                    raise webob.exc.HTTPBadRequest(explanation=msg)
            # Ensure quota usage wouldn't exceed limit on a delete
            self._validate_existing_resource(
                res, defaults[res], project_quotas)

        db.quota_destroy_by_project(ctxt, target_project.id)

        for res, limit in project_quotas.items():
            # Update child limit to 0 so the parent hierarchy gets it's
            # allocated values updated properly
            self._update_nested_quota_allocated(
                ctxt, target_project, project_quotas, res, 0)

    def validate_setup_for_nested_quota_use(self, req):
        """Validates that the setup supports using nested quotas.

        Ensures that Keystone v3 or greater is being used, and that the
        existing quotas make sense to nest in the current hierarchy (e.g. that
        no child quota would be larger than it's parent).
        """
        ctxt = req.environ['cinder.context']
        ctxt.authorize(policy.VALIDATE_NESTED_QUOTA_POLICY)
        params = req.params
        try:
            resources = QUOTAS.resources
            resources.update(GROUP_QUOTAS.resources)
            allocated = params.get('fix_allocated_quotas', 'False')
            try:
                fix_allocated = strutils.bool_from_string(allocated,
                                                          strict=True)
            except ValueError:
                msg = _("Invalid param 'fix_allocated_quotas':%s") % allocated
                raise webob.exc.HTTPBadRequest(explanation=msg)

            quota_utils.validate_setup_for_nested_quota_use(
                ctxt, resources, quota.NestedDbQuotaDriver(),
                fix_allocated_quotas=fix_allocated)
        except exception.InvalidNestedQuotaSetup as e:
            raise webob.exc.HTTPBadRequest(explanation=e.msg)


class Quotas(extensions.ExtensionDescriptor):
    """Quota management support."""

    name = "Quotas"
    alias = "os-quota-sets"
    updated = "2011-08-08T00:00:00+00:00"

    def get_resources(self):
        resources = []

        res = extensions.ResourceExtension(
            'os-quota-sets', QuotaSetsController(),
            member_actions={'defaults': 'GET'},
            collection_actions={'validate_setup_for_nested_quota_use': 'GET'})
        resources.append(res)

        return resources



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Tests for Volume usage audit feature."""

import datetime

from cinder import context
from cinder import db
from cinder import objects
from cinder.objects import fields
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import volume as base


class GetActiveByWindowTestCase(base.BaseVolumeTestCase):
    def setUp(self):
        super(GetActiveByWindowTestCase, self).setUp()
        self.ctx = context.get_admin_context(read_deleted="yes")
        self.db_vol_attrs = [
            {
                'id': fake.VOLUME_ID,
                'host': 'devstack',
                'project_id': fake.PROJECT_ID,
                'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
                'deleted': True, 'status': 'deleted',
                'deleted_at': datetime.datetime(1, 2, 1, 1, 1, 1),
                'volume_type_id': fake.VOLUME_TYPE_ID,
            },
            {
                'id': fake.VOLUME2_ID,
                'host': 'devstack',
                'project_id': fake.PROJECT_ID,
                'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
                'deleted': True, 'status': 'deleted',
                'deleted_at': datetime.datetime(1, 3, 10, 1, 1, 1),
                'volume_type_id': fake.VOLUME_TYPE_ID,
            },
            {
                'id': fake.VOLUME3_ID,
                'host': 'devstack',
                'project_id': fake.PROJECT_ID,
                'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
                'deleted': True, 'status': 'deleted',
                'deleted_at': datetime.datetime(1, 5, 1, 1, 1, 1),
                'volume_type_id': fake.VOLUME_TYPE_ID,
            },
            {
                'id': fake.VOLUME4_ID,
                'host': 'devstack',
                'project_id': fake.PROJECT_ID,
                'created_at': datetime.datetime(1, 3, 10, 1, 1, 1),
                'volume_type_id': fake.VOLUME_TYPE_ID,
            },
            {
                'id': fake.VOLUME5_ID,
                'host': 'devstack',
                'project_id': fake.PROJECT_ID,
                'created_at': datetime.datetime(1, 5, 1, 1, 1, 1),
                'volume_type_id': fake.VOLUME_TYPE_ID,
            }
        ]

        self.db_snap_attrs = [
            {
                'id': fake.SNAPSHOT_ID,
                'project_id': 'p1',
                'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
                'deleted': True,
                'status': fields.SnapshotStatus.DELETED,
                'deleted_at': datetime.datetime(1, 2, 1, 1, 1, 1),
                'volume_id': fake.VOLUME_ID,
                'volume_type_id': fake.VOLUME_TYPE_ID,
            },

            {
                'id': fake.SNAPSHOT2_ID,
                'project_id': 'p1',
                'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
                'deleted': True,
                'status': fields.SnapshotStatus.DELETED,
                'deleted_at': datetime.datetime(1, 3, 10, 1, 1, 1),
                'volume_id': fake.VOLUME_ID,
                'volume_type_id': fake.VOLUME_TYPE_ID,
            },
            {
                'id': fake.SNAPSHOT3_ID,
                'project_id': 'p1',
                'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
                'deleted': True,
                'status': fields.SnapshotStatus.DELETED,
                'deleted_at': datetime.datetime(1, 5, 1, 1, 1, 1),
                'volume_id': fake.VOLUME_ID,
                'volume_type_id': fake.VOLUME_TYPE_ID,
            },
            {
                'id': fake.SNAPSHOT_ID,
                'project_id': 'p1',
                'created_at': datetime.datetime(1, 3, 10, 1, 1, 1),
                'volume_id': fake.VOLUME_ID,
                'volume_type_id': fake.VOLUME_TYPE_ID,
            },
            {
                'id': fake.SNAPSHOT2_ID,
                'project_id': 'p1',
                'created_at': datetime.datetime(1, 5, 1, 1, 1, 1),
                'volume_id': fake.VOLUME_ID,
                'volume_type_id': fake.VOLUME_TYPE_ID
            }
        ]

        self.db_back_attrs = [
            {
                'id': fake.BACKUP_ID,
                'host': 'devstack',
                'project_id': fake.PROJECT_ID,
                'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
                'deleted': 1,
                'status': 'deleted',
                'deleted_at': datetime.datetime(1, 2, 1, 1, 1, 1)
            },
            {
                'id': fake.BACKUP2_ID,
                'host': 'devstack',
                'project_id': fake.PROJECT_ID,
                'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
                'deleted': 1,
                'status': 'deleted',
                'deleted_at': datetime.datetime(1, 3, 10, 1, 1, 1)
            },
            {
                'id': fake.BACKUP3_ID,
                'host': 'devstack',
                'project_id': fake.PROJECT_ID,
                'created_at': datetime.datetime(1, 1, 1, 1, 1, 1),
                'deleted': 1,
                'status': 'deleted',
                'deleted_at': datetime.datetime(1, 5, 1, 1, 1, 1)
            },
            {
                'id': fake.BACKUP4_ID,
                'host': 'devstack',
                'project_id': fake.PROJECT_ID,
                'created_at': datetime.datetime(1, 3, 10, 1, 1, 1),
            },
            {
                'id': fake.BACKUP5_ID,
                'host': 'devstack',
                'project_id': fake.PROJECT_ID,
                'created_at': datetime.datetime(1, 5, 1, 1, 1, 1),
            },
        ]

    def test_volume_get_all_active_by_window(self):
        # Find all all volumes valid within a timeframe window.

        # Not in window
        db.volume_create(self.ctx, self.db_vol_attrs[0])

        # In - deleted in window
        db.volume_create(self.ctx, self.db_vol_attrs[1])

        # In - deleted after window
        db.volume_create(self.ctx, self.db_vol_attrs[2])

        # In - created in window
        db.volume_create(self.context, self.db_vol_attrs[3])

        # Not of window.
        db.volume_create(self.context, self.db_vol_attrs[4])

        volumes = db.volume_get_all_active_by_window(
            self.context,
            datetime.datetime(1, 3, 1, 1, 1, 1),
            datetime.datetime(1, 4, 1, 1, 1, 1),
            project_id=fake.PROJECT_ID)
        self.assertEqual(3, len(volumes))
        self.assertEqual(fake.VOLUME2_ID, volumes[0].id)
        self.assertEqual(fake.VOLUME3_ID, volumes[1].id)
        self.assertEqual(fake.VOLUME4_ID, volumes[2].id)

    def test_snapshot_get_all_active_by_window(self):
        # Find all all snapshots valid within a timeframe window.
        db.volume_create(self.context, {'id': fake.VOLUME_ID,
                                        'volume_type_id':
                                            fake.VOLUME_TYPE_ID})
        for i in range(5):
            self.db_vol_attrs[i]['volume_id'] = fake.VOLUME_ID

        # Not in window
        del self.db_snap_attrs[0]['id']
        snap1 = objects.Snapshot(self.ctx, **self.db_snap_attrs[0])
        snap1.create()

        # In - deleted in window
        del self.db_snap_attrs[1]['id']
        snap2 = objects.Snapshot(self.ctx, **self.db_snap_attrs[1])
        snap2.create()

        # In - deleted after window
        del self.db_snap_attrs[2]['id']
        snap3 = objects.Snapshot(self.ctx, **self.db_snap_attrs[2])
        snap3.create()

        # In - created in window
        del self.db_snap_attrs[3]['id']
        snap4 = objects.Snapshot(self.ctx, **self.db_snap_attrs[3])
        snap4.create()

        # Not of window.
        del self.db_snap_attrs[4]['id']
        snap5 = objects.Snapshot(self.ctx, **self.db_snap_attrs[4])
        snap5.create()

        snapshots = objects.SnapshotList.get_all_active_by_window(
            self.context,
            datetime.datetime(1, 3, 1, 1, 1, 1),
            datetime.datetime(1, 4, 1, 1, 1, 1)).objects
        self.assertEqual(3, len(snapshots))
        self.assertEqual(snap2.id, snapshots[0].id)
        self.assertEqual(fake.VOLUME_ID, snapshots[0].volume_id)
        self.assertEqual(snap3.id, snapshots[1].id)
        self.assertEqual(fake.VOLUME_ID, snapshots[1].volume_id)
        self.assertEqual(snap4.id, snapshots[2].id)
        self.assertEqual(fake.VOLUME_ID, snapshots[2].volume_id)

    def test_backup_get_all_active_by_window(self):
        # Find all backups valid within a timeframe window.
        db.volume_create(self.context, {'id': fake.VOLUME_ID,
                                        'volume_type_id':
                                            fake.VOLUME_TYPE_ID})
        for i in range(5):
            self.db_back_attrs[i]['volume_id'] = fake.VOLUME_ID

        # Not in window
        db.backup_create(self.ctx, self.db_back_attrs[0])

        # In - deleted in window
        db.backup_create(self.ctx, self.db_back_attrs[1])

        # In - deleted after window
        db.backup_create(self.ctx, self.db_back_attrs[2])

        # In - created in window
        db.backup_create(self.ctx, self.db_back_attrs[3])

        # Not of window
        db.backup_create(self.ctx, self.db_back_attrs[4])

        backups = db.backup_get_all_active_by_window(
            self.context,
            datetime.datetime(1, 3, 1, 1, 1, 1),
            datetime.datetime(1, 4, 1, 1, 1, 1),
            project_id=fake.PROJECT_ID
        )
        self.assertEqual(3, len(backups))
        self.assertEqual(fake.BACKUP2_ID, backups[0].id)
        self.assertEqual(fake.BACKUP3_ID, backups[1].id)
        self.assertEqual(fake.BACKUP4_ID, backups[2].id)



# Copyright 2011 Justin Santa Barbara
# Copyright 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Implementation of a fake image service."""

import copy
import datetime
from unittest import mock
import uuid

from cinder import exception
import cinder.image.glance
from cinder.tests.unit import fake_constants


class _FakeImageService(object):
    """Mock (fake) image service for unit testing."""

    def __init__(self):
        self.images = {}
        # NOTE(justinsb): The OpenStack API can't upload an image?
        # So, make sure we've got one..
        timestamp = datetime.datetime(2011, 1, 1, 1, 2, 3)

        image1 = {'id': fake_constants.IMAGE_ID,
                  'name': 'fakeimage123456',
                  'created_at': timestamp,
                  'updated_at': timestamp,
                  'deleted_at': None,
                  'deleted': False,
                  'status': 'active',
                  'visibility': 'private',
                  'protected': False,
                  'container_format': 'raw',
                  'disk_format': 'raw',
                  'properties': {'kernel_id': 'nokernel',
                                 'ramdisk_id': 'nokernel',
                                 'architecture': 'x86_64'},
                  'size': 12345678}

        image2 = {'id': 'a2459075-d96c-40d5-893e-577ff92e721c',
                  'name': 'fakeimage123456',
                  'created_at': timestamp,
                  'updated_at': timestamp,
                  'deleted_at': None,
                  'deleted': False,
                  'status': 'active',
                  'visibility': 'public',
                  'protected': True,
                  'container_format': 'ami',
                  'disk_format': 'ami',
                  'properties': {'kernel_id': 'nokernel',
                                 'ramdisk_id': 'nokernel'},
                  'size': 1}

        image3 = {'id': '76fa36fc-c930-4bf3-8c8a-ea2a2420deb6',
                  'name': 'fakeimage123456',
                  'created_at': timestamp,
                  'updated_at': timestamp,
                  'deleted_at': None,
                  'deleted': False,
                  'status': 'active',
                  'visibility': 'public',
                  'protected': True,
                  'container_format': None,
                  'disk_format': None,
                  'properties': {'kernel_id': 'nokernel',
                                 'ramdisk_id': 'nokernel'},
                  'size': 1000000000000}

        image4 = {'id': 'cedef40a-ed67-4d10-800e-17455edce175',
                  'name': 'fakeimage123456',
                  'created_at': timestamp,
                  'updated_at': timestamp,
                  'deleted_at': None,
                  'deleted': False,
                  'status': 'active',
                  'visibility': 'public',
                  'protected': True,
                  'container_format': 'ami',
                  'disk_format': 'ami',
                  'properties': {'kernel_id': 'nokernel',
                                 'ramdisk_id': 'nokernel'},
                  'size': 20000000}

        image5 = {'id': 'c905cedb-7281-47e4-8a62-f26bc5fc4c77',
                  'name': 'fakeimage123456',
                  'created_at': timestamp,
                  'updated_at': timestamp,
                  'deleted_at': None,
                  'deleted': False,
                  'status': 'active',
                  'visibility': 'public',
                  'protected': True,
                  'container_format': 'ami',
                  'disk_format': 'ami',
                  'properties': {
                      'kernel_id': '155d900f-4e14-4e4c-a73d-069cbf4541e6',
                      'ramdisk_id': None},
                  'size': 50000}

        image6 = {'id': 'a440c04b-79fa-479c-bed1-0b816eaec379',
                  'name': 'fakeimage6',
                  'created_at': timestamp,
                  'updated_at': timestamp,
                  'deleted_at': None,
                  'deleted': False,
                  'status': 'active',
                  'visibility': 'public',
                  'protected': False,
                  'container_format': 'ova',
                  'disk_format': 'vhd',
                  'properties': {'kernel_id': 'nokernel',
                                 'ramdisk_id': 'nokernel',
                                 'architecture': 'x86_64',
                                 'auto_disk_config': 'False'},
                  'size': 7777777}

        image7 = {'id': '70a599e0-31e7-49b7-b260-868f441e862b',
                  'name': 'fakeimage7',
                  'created_at': timestamp,
                  'updated_at': timestamp,
                  'deleted_at': None,
                  'deleted': False,
                  'status': 'active',
                  'visibility': 'public',
                  'protected': False,
                  'container_format': 'ova',
                  'disk_format': 'vhd',
                  'properties': {'kernel_id': 'nokernel',
                                 'ramdisk_id': 'nokernel',
                                 'architecture': 'x86_64',
                                 'auto_disk_config': 'True'},
                  'size': 1234000000}

        image8 = {'id': 'b0a599e0-41d7-3582-b260-769f443c862a',
                  'name': 'fakeimage8',
                  'created_at': timestamp,
                  'updated_at': timestamp,
                  'deleted_at': None,
                  'deleted': False,
                  'status': 'active',
                  'is_public': False,
                  'container_format': 'bare',
                  'disk_format': 'raw',
                  'properties':
                  {'block_device_mapping': [
                      {'boot_index': 0, 'source_type': 'snapshot',
                       'snapshot_id': fake_constants.SNAPSHOT_ID}],
                      'ramdisk_id': 'nokernel',
                      'architecture': 'x86_64',
                      'auto_disk_config': 'True'}}

        self.create(None, image1)
        self.create(None, image2)
        self.create(None, image3)
        self.create(None, image4)
        self.create(None, image5)
        self.create(None, image6)
        self.create(None, image7)
        self.create(None, image8)
        self._imagedata = {}
        self.temp_images = mock.MagicMock()
        super(_FakeImageService, self).__init__()

    # TODO(bcwaldon): implement optional kwargs such as limit, sort_dir
    def detail(self, context, **kwargs):
        """Return list of detailed image information."""
        return copy.deepcopy(self.images.values())

    def download(self, context, image_id, data):
        self.show(context, image_id)
        data.write(self._imagedata.get(image_id, ''))

    def show(self, context, image_id):
        """Get data about specified image.

        Returns a dict containing image data for the given opaque image id.

        """
        image = self.images.get(str(image_id))
        if image:
            return copy.deepcopy(image)
        raise exception.ImageNotFound(image_id=image_id)

    def create(self, context, metadata, data=None):
        """Store the image data and return the new image id.

        :raises Duplicate: if the image already exist.

        """
        image_id = str(metadata.get('id', uuid.uuid4()))
        metadata['id'] = image_id
        if image_id in self.images:
            raise exception.Duplicate()
        self.images[image_id] = copy.deepcopy(metadata)
        if data:
            self._imagedata[image_id] = data.read()
        return self.images[image_id]

    def update(self, context, image_id, metadata, data=None,
               purge_props=False, store_id=None, base_image_ref=None):
        """Replace the contents of the given image with the new data.

        :raises ImageNotFound: if the image does not exist.

        """
        if not self.images.get(image_id):
            raise exception.ImageNotFound(image_id=image_id)
        if purge_props:
            self.images[image_id] = copy.deepcopy(metadata)
        else:
            image = self.images[image_id]
            try:
                image['properties'].update(metadata.pop('properties'))
            except Exception:
                pass
            image.update(metadata)
        return self.images[image_id]

    def delete(self, context, image_id):
        """Delete the given image.

        :raises ImageNotFound: if the image does not exist.

        """
        removed = self.images.pop(image_id, None)
        if not removed:
            raise exception.ImageNotFound(image_id=image_id)

    def get_location(self, context, image_id):
        if image_id in self.images:
            return 'fake_location'
        return None

    def add_location(self, context, image_id, url, metadata):
        self.update(context, image_id, {'locations': [{'url': url,
                                                       'metadata': metadata}]})
        return True


_fakeImageService = _FakeImageService()


def FakeImageService():
    return _fakeImageService


def FakeImageService_reset():
    global _fakeImageService
    _fakeImageService = _FakeImageService()


def mock_image_service(testcase):
    testcase.mock_object(cinder.image.glance, 'get_remote_image_service',
                         lambda x, y: (FakeImageService(), y))
    testcase.mock_object(cinder.image.glance, 'get_default_image_service',
                         mock.Mock(side_effect=FakeImageService))



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Utilities for Dell EMC PowerStore Cinder driver."""

import re

from oslo_log import log as logging
from oslo_utils import units

from cinder import exception
from cinder.i18n import _
from cinder.objects import fields


LOG = logging.getLogger(__name__)


def bytes_to_gib(size_in_bytes):
    """Convert size in bytes to GiB.

    :param size_in_bytes: size in bytes
    :return: size in GiB
    """

    return size_in_bytes // units.Gi


def gib_to_bytes(size_in_gb):
    """Convert size in GiB to bytes.

    :param size_in_gb: size in GiB
    :return: size in bytes
    """

    return size_in_gb * units.Gi


def extract_fc_wwpns(connector):
    """Convert connector FC ports to appropriate format with colons.

    :param connector: connection properties
    :return: FC ports in appropriate format with colons
    """

    if "wwnns" not in connector or "wwpns" not in connector:
        msg = _("Host %s does not have FC initiators.") % connector["host"]
        LOG.error(msg)
        raise exception.VolumeBackendAPIException(data=msg)
    return [":".join(re.findall("..", wwpn)) for wwpn in connector["wwpns"]]


def fc_wwn_to_string(wwn):
    """Convert FC WWN to string without colons.

    :param wwn: FC WWN
    :return: FC WWN without colons
    """

    return wwn.replace(":", "")


def iscsi_portal_with_port(address):
    """Add default port 3260 to iSCSI portal

    :param address: iSCSI portal without port
    :return: iSCSI portal with default port 3260
    """

    return "%(address)s:3260" % {"address": address}


def powerstore_host_name(connector, protocol):
    """Generate PowerStore host name for connector.

    :param connector: connection properties
    :param protocol: storage protocol (FC or iSCSI)
    :return: unique host name
    """

    return ("%(host)s-%(protocol)s" %
            {"host": connector["host"],
             "protocol": protocol, })


def filter_hosts_by_initiators(hosts, initiators):
    """Filter hosts by given list of initiators.

    :param hosts: list of PowerStore host objects
    :param initiators: list of initiators
    :return: PowerStore hosts list
    """

    hosts_names_found = set()
    for host in hosts:
        for initiator in host["host_initiators"]:
            if initiator["port_name"] in initiators:
                hosts_names_found.add(host["name"])
    return list(filter(lambda host: host["name"] in hosts_names_found, hosts))


def is_multiattached_to_host(volume_attachment, host_name):
    """Check if volume is attached to multiple instances on one host.

    When multiattach is enabled, a volume could be attached to two or more
    instances which are hosted on one nova host.
    Because PowerStore cannot recognize the volume is attached to two or more
    instances, we should keep the volume attached to the nova host until
    the volume is detached from the last instance.

    :param volume_attachment: list of VolumeAttachment objects
    :param host_name: OpenStack host name
    :return: multiattach flag
    """

    if not volume_attachment:
        return False

    attachments = [
        attachment for attachment in volume_attachment
        if (attachment.attach_status == fields.VolumeAttachStatus.ATTACHED and
            attachment.attached_host == host_name)
    ]
    return len(attachments) > 1



# Copyright 2018 Red Hat, Inc
# Copyright 2017 Rackspace Australia
# Copyright 2018 Michael Still and Aptira
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Helpers for path related routines.
"""

import os

from cinder import exception
import cinder.privsep


@cinder.privsep.sys_admin_pctxt.entrypoint
def readfile(path):
    if not os.path.exists(path):
        raise exception.FileNotFound(file_path=path)
    with open(path, 'r') as f:
        return f.read()


@cinder.privsep.sys_admin_pctxt.entrypoint
def removefile(path):
    if not os.path.exists(path):
        raise exception.FileNotFound(file_path=path)
    os.unlink(path)


@cinder.privsep.sys_admin_pctxt.entrypoint
def touch(path):
    if os.path.exists(path):
        os.utime(path, None)
    else:
        open(path, 'a').close()


@cinder.privsep.sys_admin_pctxt.entrypoint
def symlink(src, dest):
    if not os.path.exists(src):
        raise exception.FileNotFound(file_path=src)
    os.symlink(src, dest)



# Copyright (c) 2016 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import unittest
from unittest import mock

from oslo_utils import units

from cinder import exception
from cinder.tests.unit.volume.drivers.dell_emc.unity import test_adapter
from cinder.tests.unit.volume.drivers.dell_emc.unity import test_driver
from cinder.volume.drivers.dell_emc.unity import utils


def get_volume_type_extra_specs(volume_type):
    return {'provisioning:type': volume_type}


def get_group_type_specs(group_type):
    return {'consistent_group_snapshot_enabled': '<is> True',
            'group_type_id': group_type}


def get_volume_type_qos_specs(type_id):
    if type_id == 'invalid_backend_qos_consumer':
        ret = {'qos_specs': {'consumer': 'invalid'}}
    elif type_id == 'both_none':
        ret = {'qos_specs': {'consumer': 'back-end', 'specs': {}}}
    elif type_id == 'max_1000_iops':
        ret = {
            'qos_specs': {
                'id': 'max_1000_iops',
                'consumer': 'both',
                'specs': {
                    'maxIOPS': 1000
                }
            }
        }
    elif type_id == 'max_2_mbps':
        ret = {
            'qos_specs': {
                'id': 'max_2_mbps',
                'consumer': 'back-end',
                'specs': {
                    'maxBWS': 2
                }
            }
        }
    else:
        ret = None
    return ret


def patch_volume_types(func):
    @functools.wraps(func)
    @mock.patch(target=('cinder.volume.volume_types'
                        '.get_volume_type_extra_specs'),
                new=get_volume_type_extra_specs)
    @mock.patch(target=('cinder.volume.volume_types'
                        '.get_volume_type_qos_specs'),
                new=get_volume_type_qos_specs)
    def func_wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return func_wrapper


def patch_group_types(func):
    @functools.wraps(func)
    @mock.patch(target=('cinder.volume.group_types'
                        '.get_group_type_specs'),
                new=get_group_type_specs)
    def func_wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return func_wrapper


class UnityUtilsTest(unittest.TestCase):
    def test_validate_pool_names_filter(self):
        all_pools = list('acd')
        pool_names = utils.validate_pool_names(list('abc'), all_pools)
        self.assertIn('a', pool_names)
        self.assertIn('c', pool_names)
        self.assertNotIn('b', pool_names)
        self.assertNotIn('d', pool_names)

    def test_validate_pool_names_non_exists(self):
        def f():
            all_pools = list('abc')
            utils.validate_pool_names(list('efg'), all_pools)

        self.assertRaises(exception.VolumeBackendAPIException, f)

    def test_validate_pool_names_default(self):
        all_pools = list('ab')
        pool_names = utils.validate_pool_names([], all_pools)
        self.assertEqual(2, len(pool_names))

        pool_names = utils.validate_pool_names(None, all_pools)
        self.assertEqual(2, len(pool_names))

    def test_build_provider_location(self):
        location = utils.build_provider_location('unity', 'thin', 'ev_1', '3')
        expected = 'id^ev_1|system^unity|type^thin|version^3'
        self.assertEqual(expected, location)

    def test_extract_provider_location_version(self):
        location = 'id^ev_1|system^unity|type^thin|version^3'
        self.assertEqual('3',
                         utils.extract_provider_location(location, 'version'))

    def test_extract_provider_location_type(self):
        location = 'id^ev_1|system^unity|type^thin|version^3'
        self.assertEqual('thin',
                         utils.extract_provider_location(location, 'type'))

    def test_extract_provider_location_system(self):
        location = 'id^ev_1|system^unity|type^thin|version^3'
        self.assertEqual('unity',
                         utils.extract_provider_location(location, 'system'))

    def test_extract_provider_location_id(self):
        location = 'id^ev_1|system^unity|type^thin|version^3'
        self.assertEqual('ev_1',
                         utils.extract_provider_location(location, 'id'))

    def test_extract_provider_location_not_found(self):
        location = 'id^ev_1|system^unity|type^thin|version^3'
        self.assertIsNone(utils.extract_provider_location(location, 'na'))

    def test_extract_provider_location_none(self):
        self.assertIsNone(utils.extract_provider_location(None, 'abc'))

    def test_extract_iscsi_uids(self):
        connector = {'host': 'fake_host',
                     'initiator': 'fake_iqn'}
        self.assertEqual(['fake_iqn'],
                         utils.extract_iscsi_uids(connector))

    def test_extract_iscsi_uids_not_found(self):
        connector = {'host': 'fake_host'}
        self.assertRaises(exception.VolumeBackendAPIException,
                          utils.extract_iscsi_uids,
                          connector)

    def test_extract_fc_uids(self):
        connector = {'host': 'fake_host',
                     'wwnns': ['1111111111111111',
                               '2222222222222222'],
                     'wwpns': ['3333333333333333',
                               '4444444444444444']
                     }
        self.assertEqual(['11:11:11:11:11:11:11:11:33:33:33:33:33:33:33:33',
                          '22:22:22:22:22:22:22:22:44:44:44:44:44:44:44:44', ],
                         utils.extract_fc_uids(connector))

    def test_extract_fc_uids_not_found(self):
        connector = {'host': 'fake_host'}
        self.assertRaises(exception.VolumeBackendAPIException,
                          utils.extract_iscsi_uids,
                          connector)

    def test_byte_to_gib(self):
        self.assertEqual(5, utils.byte_to_gib(5 * units.Gi))

    def test_byte_to_mib(self):
        self.assertEqual(5, utils.byte_to_mib(5 * units.Mi))

    def test_gib_to_mib(self):
        self.assertEqual(5 * units.Gi / units.Mi, utils.gib_to_mib(5))

    def test_convert_ip_to_portal(self):
        self.assertEqual('1.2.3.4:3260', utils.convert_ip_to_portal('1.2.3.4'))
        self.assertEqual('[fd27:2e95:e174::100]:3260',
                         utils.convert_ip_to_portal('fd27:2e95:e174::100'))
        self.assertEqual('[fd27:2e95:e174::100]:3260',
                         utils.convert_ip_to_portal('[fd27:2e95:e174::100]'))

    def test_convert_to_itor_tgt_map(self):
        zone_mapping = {
            'san_1': {
                'initiator_port_wwn_list':
                    ('200000051e55a100', '200000051e55a121'),
                'target_port_wwn_list':
                    ('100000051e55a100', '100000051e55a121')
            }
        }
        ret = utils.convert_to_itor_tgt_map(zone_mapping)
        self.assertEqual(['100000051e55a100', '100000051e55a121'], ret[0])
        mapping = ret[1]
        targets = ('100000051e55a100', '100000051e55a121')
        self.assertEqual(targets, mapping['200000051e55a100'])
        self.assertEqual(targets, mapping['200000051e55a121'])

    def test_get_pool_name(self):
        volume = test_adapter.MockOSResource(host='host@backend#pool_name')
        self.assertEqual('pool_name', utils.get_pool_name(volume))

    def test_get_pool_name_from_host(self):
        host = {'host': 'host@backend#pool_name'}
        ret = utils.get_pool_name_from_host(host)
        self.assertEqual('pool_name', ret)

    def get_backend_name_from_volume(self):
        volume = test_adapter.MockOSResource(host='host@backend#pool_name')
        ret = utils.get_backend_name_from_volume(volume)
        self.assertEqual('host@backend', ret)

    def get_backend_name_from_host(self):
        host = {'host': 'host@backend#pool_name'}
        ret = utils.get_backend_name_from_volume(host)
        self.assertEqual('host@backend', ret)

    def test_ignore_exception(self):
        class IgnoredException(Exception):
            pass

        def f():
            raise IgnoredException('any exception')

        try:
            utils.ignore_exception(f)
        except IgnoredException:
            self.fail('should not raise any exception.')

    def test_assure_cleanup(self):
        data = [0]

        def _enter():
            data[0] += 10
            return data[0]

        def _exit(x):
            data[0] = x - 1

        ctx = utils.assure_cleanup(_enter, _exit, True)
        with ctx as r:
            self.assertEqual(10, r)

        self.assertEqual(9, data[0])

    def test_get_backend_qos_specs_type_none(self):
        volume = test_adapter.MockOSResource(volume_type_id=None)
        ret = utils.get_backend_qos_specs(volume)
        self.assertIsNone(ret)

    @patch_volume_types
    def test_get_backend_qos_specs_none(self):
        volume = test_adapter.MockOSResource(volume_type_id='no_qos')
        ret = utils.get_backend_qos_specs(volume)
        self.assertIsNone(ret)

    @patch_volume_types
    def test_get_backend_qos_invalid_consumer(self):
        volume = test_adapter.MockOSResource(
            volume_type_id='invalid_backend_qos_consumer')
        ret = utils.get_backend_qos_specs(volume)
        self.assertIsNone(ret)

    @patch_volume_types
    def test_get_backend_qos_both_none(self):
        volume = test_adapter.MockOSResource(volume_type_id='both_none')
        ret = utils.get_backend_qos_specs(volume)
        self.assertIsNone(ret)

    @patch_volume_types
    def test_get_backend_qos_iops(self):
        volume = test_adapter.MockOSResource(volume_type_id='max_1000_iops')
        ret = utils.get_backend_qos_specs(volume)
        expected = {'maxBWS': None, 'id': 'max_1000_iops', 'maxIOPS': 1000}
        self.assertEqual(expected, ret)

    @patch_volume_types
    def test_get_backend_qos_mbps(self):
        volume = test_adapter.MockOSResource(volume_type_id='max_2_mbps')
        ret = utils.get_backend_qos_specs(volume)
        expected = {'maxBWS': 2, 'id': 'max_2_mbps', 'maxIOPS': None}
        self.assertEqual(expected, ret)

    def test_remove_empty(self):
        option = mock.Mock()
        value_list = [' pool1', 'pool2 ', '     pool3  ']
        ret = utils.remove_empty(option, value_list)
        expected = ['pool1', 'pool2', 'pool3']
        self.assertListEqual(expected, ret)

    def test_remove_empty_none(self):
        option = mock.Mock()
        value_list = None
        ret = utils.remove_empty(option, value_list)
        expected = None
        self.assertEqual(expected, ret)

    def test_remove_empty_empty_list(self):
        option = mock.Mock()
        value_list = []
        ret = utils.remove_empty(option, value_list)
        expected = None
        self.assertEqual(expected, ret)

    @patch_group_types
    def test_group_is_cg(self):
        cg = test_driver.UnityDriverTest.get_cg()
        result = utils.group_is_cg(cg)
        self.assertTrue(result)

    @patch_group_types
    def test_get_group_specs_by_key(self):
        cg = test_driver.UnityDriverTest.get_cg()
        result = utils.get_group_specs(cg, 'consistent_group_snapshot_enabled')
        self.assertEqual('<is> True', result)

    @patch_group_types
    def test_no_group_specs_key(self):
        cg = test_driver.UnityDriverTest.get_cg()
        result = utils.get_group_specs(cg, 'test_key')
        self.assertIsNone(result)

    @patch_volume_types
    def test_retype_no_need_migration_when_same_host(self):
        volume = test_adapter.MockOSResource(volume_type_id='host_1',
                                             host='host_1')
        new_host = {'name': 'new_name', 'host': 'host_1'}
        ret = utils.retype_need_migration(volume, None, None, new_host)
        self.assertFalse(ret)

    @patch_volume_types
    def test_retype_need_migration_when_diff_host(self):
        volume = test_adapter.MockOSResource(volume_type_id='host_1',
                                             host='host_1')
        new_host = {'name': 'new_name', 'host': 'new_host'}
        ret = utils.retype_need_migration(volume, None, None, new_host)
        self.assertTrue(ret)

    @patch_volume_types
    def test_retype_no_need_migration_thin_to_compressed(self):
        volume = test_adapter.MockOSResource(volume_type_id='host_1',
                                             host='host_1')
        new_host = {'name': 'new_name', 'host': 'host_1'}
        old_provision = ''
        new_provision = 'compressed'
        ret = utils.retype_need_migration(volume, old_provision,
                                          new_provision, new_host)
        self.assertFalse(ret)

    @patch_volume_types
    def test_retype_no_need_migration_compressed_to_thin(self):
        volume = test_adapter.MockOSResource(volume_type_id='host_1',
                                             host='host_1')
        new_host = {'name': 'new_name', 'host': 'host_1'}
        old_provision = 'compressed'
        new_provision = ''
        ret = utils.retype_need_migration(volume, old_provision,
                                          new_provision, new_host)
        self.assertFalse(ret)

    @patch_volume_types
    def test_retype_need_migration_thin_to_thick(self):
        volume = test_adapter.MockOSResource(volume_type_id='host_1',
                                             host='host_1')
        new_host = {'name': 'new_name', 'host': 'host_1'}
        old_provision = ''
        new_provision = 'thick'
        ret = utils.retype_need_migration(volume, old_provision,
                                          new_provision, new_host)
        self.assertTrue(ret)



# Copyright 2016 Red Hat, Inc
# Copyright 2017 Rackspace Australia
# Copyright 2018 Michael Still and Aptira
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Helpers for cgroup related routines.
"""

from oslo_concurrency import processutils

import cinder.privsep


@cinder.privsep.sys_admin_pctxt.entrypoint
def cgroup_create(name):
    processutils.execute('cgcreate', '-g', 'blkio:%s' % name)


@cinder.privsep.sys_admin_pctxt.entrypoint
def cgroup_limit(name, rw, dev, bps):
    processutils.execute('cgset', '-r',
                         'blkio.throttle.%s_bps_device=%s %d' % (rw, dev, bps),
                         name)



#    (c) Copyright 2012-2015 Hewlett Packard Enterprise Development LP
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
"""
Volume driver for HPE 3PAR Storage array.
This driver requires 3.1.3 or later firmware on the 3PAR array, using
the 4.x version of the hpe3parclient.

You will need to install the python hpe3parclient.
sudo pip install --upgrade "hpe3parclient>=4.0"

Set the following in the cinder.conf file to enable the
3PAR iSCSI Driver along with the required flags:

volume_driver=cinder.volume.drivers.hpe.hpe_3par_iscsi.HPE3PARISCSIDriver
"""

import re
import sys

try:
    from hpe3parclient import exceptions as hpeexceptions
except ImportError:
    hpeexceptions = None

from oslo_log import log as logging
from oslo_utils.excutils import save_and_reraise_exception

from cinder import coordination
from cinder import exception
from cinder.i18n import _
from cinder import interface
from cinder import utils
from cinder.volume.drivers.hpe import hpe_3par_base as hpebasedriver
from cinder.volume import volume_utils

LOG = logging.getLogger(__name__)

# EXISTENT_PATH error code returned from hpe3parclient
EXISTENT_PATH = 73
DEFAULT_ISCSI_PORT = 3260
CHAP_USER_KEY = "HPQ-cinder-CHAP-name"
CHAP_PASS_KEY = "HPQ-cinder-CHAP-secret"


@interface.volumedriver
class HPE3PARISCSIDriver(hpebasedriver.HPE3PARDriverBase):
    """OpenStack iSCSI driver to enable 3PAR storage array.

    Version history:

    .. code-block:: none

        1.0   - Initial driver
        1.1   - QoS, extend volume, multiple iscsi ports, remove domain,
                session changes, faster clone, requires 3.1.2 MU2 firmware.
        1.2.0 - Updated the use of the hp3parclient to 2.0.0 and refactored
                the drivers to use the new APIs.
        1.2.1 - Synchronized extend_volume method.
        1.2.2 - Added try/finally around client login/logout.
        1.2.3 - log exceptions before raising
        1.2.4 - Fixed iSCSI active path bug #1224594
        1.2.5 - Added metadata during attach/detach bug #1258033
        1.2.6 - Use least-used iscsi n:s:p for iscsi volume attach bug #1269515
                This update now requires 3.1.2 MU3 firmware
        1.3.0 - Removed all SSH code.  We rely on the hp3parclient now.
        2.0.0 - Update hp3parclient API uses 3.0.x
        2.0.2 - Add back-end assisted volume migrate
        2.0.3 - Added support for managing/unmanaging of volumes
        2.0.4 - Added support for volume retype
        2.0.5 - Added CHAP support, requires 3.1.3 MU1 firmware
                and hp3parclient 3.1.0.
        2.0.6 - Fixing missing login/logout around attach/detach bug #1367429
        2.0.7 - Add support for pools with model update
        2.0.8 - Migrate without losing type settings bug #1356608
        2.0.9 - Removing locks bug #1381190
        2.0.10 - Add call to queryHost instead SSH based findHost #1398206
        2.0.11 - Added missing host name during attach fix #1398206
        2.0.12 - Removed usage of host name cache #1398914
        2.0.13 - Update LOG usage to fix translations.  bug #1384312
        2.0.14 - Do not allow a different iSCSI IP (hp3par_iscsi_ips) to be
                 used during live-migration.  bug #1423958
        2.0.15 - Added support for updated detach_volume attachment.
        2.0.16 - Added encrypted property to initialize_connection #1439917
        2.0.17 - Python 3 fixes
        2.0.18 - Improved VLUN creation and deletion logic. #1469816
        2.0.19 - Changed initialize_connection to use getHostVLUNs. #1475064
        2.0.20 - Adding changes to support 3PAR iSCSI multipath.
        2.0.21 - Adds consistency group support
        2.0.22 - Update driver to use ABC metaclasses
        2.0.23 - Added update_migrated_volume. bug # 1492023
        3.0.0 - Rebranded HP to HPE.
        3.0.1 - Python 3 support
        3.0.2 - Remove db access for consistency groups
        3.0.3 - Fix multipath dictionary key error. bug #1522062
        3.0.4 - Adds v2 managed replication support
        3.0.5 - Adds v2 unmanaged replication support
        3.0.6 - Adding manage/unmanage snapshot support
        3.0.7 - Optimize array ID retrieval
        3.0.8 - Update replication to version 2.1
        3.0.9 - Use same LUN ID for each VLUN path #1551994
        3.0.10 - Remove metadata that tracks the instance ID. bug #1572665
        3.0.11 - _create_3par_iscsi_host() now accepts iscsi_iqn as list only.
                 Bug #1590180
        3.0.12 - Added entry point tracing
        3.0.13 - Handling HTTP conflict 409, host WWN/iSCSI name already used
                by another host, while creating 3PAR iSCSI Host. bug #1642945
        3.0.14 - Handle manage and unmanage hosts present. bug #1648067
        3.0.15 - Adds consistency group capability in generic volume groups.
        3.0.16 - Get host from os-brick connector. bug #1690244
        4.0.0 - Adds base class.
        4.0.1 - Update CHAP on host record when volume is migrated
                to new compute host. bug # 1737181
        4.0.2 - Handle force detach case. bug #1686745
        4.0.3 - Set proper backend on subsequent operation, after group
                failover. bug #1773069
        4.0.4 - Added Peer Persistence feature
        4.0.5 - Added Primera array check. bug #1849525

    """

    VERSION = "4.0.5"

    # The name of the CI wiki page.
    CI_WIKI_NAME = "HPE_Storage_CI"

    def __init__(self, *args, **kwargs):
        super(HPE3PARISCSIDriver, self).__init__(*args, **kwargs)
        self.protocol = 'iSCSI'

    def _do_setup(self, common):
        client_obj = common.client
        is_primera = client_obj.is_primera_array()
        if is_primera:
            LOG.error("For Primera, only FC is supported. "
                      "iSCSI cannot be used")
            raise NotImplementedError()

        self.iscsi_ips = {}
        common.client_login()
        try:
            self.initialize_iscsi_ports(common)
        finally:
            self._logout(common)

    def initialize_iscsi_ports(self, common,
                               remote_target=None, remote_client=None):
        # map iscsi_ip-> ip_port
        #             -> iqn
        #             -> nsp
        iscsi_ip_list = {}
        temp_iscsi_ip = {}

        if remote_target:
            backend_conf = remote_target
        else:
            backend_conf = common._client_conf

        # use the 3PAR ip_addr list for iSCSI configuration
        if len(backend_conf['hpe3par_iscsi_ips']) > 0:
            # add port values to ip_addr, if necessary
            for ip_addr in backend_conf['hpe3par_iscsi_ips']:
                ip = ip_addr.split(':')
                if len(ip) == 1:
                    temp_iscsi_ip[ip_addr] = {'ip_port': DEFAULT_ISCSI_PORT}
                elif len(ip) == 2:
                    temp_iscsi_ip[ip[0]] = {'ip_port': ip[1]}
                else:
                    LOG.warning("Invalid IP address format '%s'", ip_addr)

        # add the single value iscsi_ip_address option to the IP dictionary.
        # This way we can see if it's a valid iSCSI IP. If it's not valid,
        # we won't use it and won't bother to report it, see below
        if 'iscsi_ip_address' in backend_conf:
            if (backend_conf['iscsi_ip_address'] not in temp_iscsi_ip):
                ip = backend_conf['iscsi_ip_address']
                ip_port = backend_conf['iscsi_port']
                temp_iscsi_ip[ip] = {'ip_port': ip_port}

        # get all the valid iSCSI ports from 3PAR
        # when found, add the valid iSCSI ip, ip port, iqn and nsp
        # to the iSCSI IP dictionary
        iscsi_ports = common.get_active_iscsi_target_ports(remote_client)

        for port in iscsi_ports:
            ip = port['IPAddr']
            if ip in temp_iscsi_ip:
                ip_port = temp_iscsi_ip[ip]['ip_port']
                iscsi_ip_list[ip] = {'ip_port': ip_port,
                                     'nsp': port['nsp'],
                                     'iqn': port['iSCSIName']}
                del temp_iscsi_ip[ip]

        # if the single value iscsi_ip_address option is still in the
        # temp dictionary it's because it defaults to $my_ip which doesn't
        # make sense in this context. So, if present, remove it and move on.
        if 'iscsi_ip_address' in backend_conf:
            if backend_conf['iscsi_ip_address'] in temp_iscsi_ip:
                del temp_iscsi_ip[backend_conf['iscsi_ip_address']]

        # lets see if there are invalid iSCSI IPs left in the temp dict
        if len(temp_iscsi_ip) > 0:
            LOG.warning("Found invalid iSCSI IP address(s) in "
                        "configuration option(s) hpe3par_iscsi_ips or "
                        "target_ip_address '%s.'",
                        (", ".join(temp_iscsi_ip)))

        if not len(iscsi_ip_list):
            msg = _('At least one valid iSCSI IP address must be set.')
            LOG.error(msg)
            raise exception.InvalidInput(reason=msg)

        if remote_target:
            self.iscsi_ips[remote_target['hpe3par_api_url']] = iscsi_ip_list
        else:
            self.iscsi_ips[common._client_conf['hpe3par_api_url']] = (
                iscsi_ip_list)

    def _initialize_connection_common(self, volume, connector, common,
                                      host, iscsi_ips, ready_ports,
                                      target_portals, target_iqns, target_luns,
                                      remote_client=None):

        # Target portal ips are defined in cinder.conf.
        target_portal_ips = iscsi_ips.keys()

        # Collect all existing VLUNs for this volume/host combination.
        existing_vluns = common.find_existing_vluns(volume, host,
                                                    remote_client)

        # Cycle through each ready iSCSI port and determine if a new
        # VLUN should be created or an existing one used.
        lun_id = None
        for port in ready_ports:
            iscsi_ip = port['IPAddr']
            if iscsi_ip in target_portal_ips:
                vlun = None
                # check for an already existing VLUN matching the
                # nsp for this iSCSI IP. If one is found, use it
                # instead of creating a new VLUN.
                for v in existing_vluns:
                    portPos = common.build_portPos(
                        iscsi_ips[iscsi_ip]['nsp'])
                    if v['portPos'] == portPos:
                        vlun = v
                        break
                else:
                    vlun = common.create_vlun(
                        volume, host, iscsi_ips[iscsi_ip]['nsp'],
                        lun_id=lun_id, remote_client=remote_client)

                    # We want to use the same LUN ID for every port
                    if lun_id is None:
                        lun_id = vlun['lun']

                iscsi_ip_port = "%s:%s" % (
                    iscsi_ip, iscsi_ips[iscsi_ip]['ip_port'])
                target_portals.append(iscsi_ip_port)
                target_iqns.append(port['iSCSIName'])
                target_luns.append(vlun['lun'])
            else:
                LOG.warning("iSCSI IP: '%s' was not found in "
                            "hpe3par_iscsi_ips list defined in "
                            "cinder.conf.", iscsi_ip)

    @utils.trace
    @coordination.synchronized('3par-{volume.id}')
    def initialize_connection(self, volume, connector):
        """Assigns the volume to a server.

        Assign any created volume to a compute node/host so that it can be
        used from that host.

        This driver returns a driver_volume_type of 'iscsi'.
        The format of the driver data is defined in _get_iscsi_properties.
        Example return value:

        .. code-block:: default

            {
                'driver_volume_type': 'iscsi',
                'data': {
                    'encrypted': False,
                    'target_discovered': True,
                    'target_iqn': 'iqn.2010-10.org.openstack:volume-00000001',
                    'target_protal': '127.0.0.1:3260',
                    'volume_id': 1,
                }
            }

        Steps to export a volume on 3PAR
          * Get the 3PAR iSCSI iqn
          * Create a host on the 3par
          * create vlun on the 3par
        """
        LOG.debug("volume id: %(volume_id)s",
                  {'volume_id': volume['id']})
        array_id = self.get_volume_replication_driver_data(volume)
        common = self._login(array_id=array_id)
        try:
            # If the volume has been failed over, we need to reinitialize
            # iSCSI ports so they represent the new array.
            if volume.get('replication_status') == 'failed-over' and (
               common._client_conf['hpe3par_api_url'] not in self.iscsi_ips):
                self.initialize_iscsi_ports(common)

            # Grab the correct iSCSI ports
            iscsi_ips = self.iscsi_ips[common._client_conf['hpe3par_api_url']]

            # we have to make sure we have a host
            host, username, password, cpg = self._create_host(
                common,
                volume,
                connector)

            multipath = connector.get('multipath')
            LOG.debug("multipath: %(multipath)s",
                      {'multipath': multipath})
            if multipath:
                ready_ports = common.client.getiSCSIPorts(
                    state=common.client.PORT_STATE_READY)

                target_portals = []
                target_iqns = []
                target_luns = []

                self._initialize_connection_common(
                    volume, connector, common,
                    host, iscsi_ips, ready_ports,
                    target_portals, target_iqns, target_luns)

                if volume.get('replication_status') == 'enabled':
                    LOG.debug('This is a replication setup')

                    remote_target = common._replication_targets[0]
                    replication_mode = remote_target['replication_mode']
                    quorum_witness_ip = (
                        remote_target.get('quorum_witness_ip'))

                    if replication_mode == 1:
                        LOG.debug('replication_mode is sync')
                        if quorum_witness_ip:
                            LOG.debug('quorum_witness_ip is present')
                            LOG.debug('Peer Persistence has been configured')
                        else:
                            LOG.debug('Since quorum_witness_ip is absent, '
                                      'considering this as Active/Passive '
                                      'replication')
                    else:
                        LOG.debug('Active/Passive replication has been '
                                  'configured')

                    if replication_mode == 1 and quorum_witness_ip:
                        remote_client = (
                            common._create_replication_client(remote_target))

                        self.initialize_iscsi_ports(
                            common, remote_target, remote_client)
                        remote_iscsi_ips = (
                            self.iscsi_ips[remote_target['hpe3par_api_url']])

                        # we have to make sure we have a host
                        host, username, password, cpg = (
                            self._create_host(
                                common, volume, connector,
                                remote_target, cpg, remote_client))

                        ready_ports = remote_client.getiSCSIPorts(
                            state=remote_client.PORT_STATE_READY)

                        self._initialize_connection_common(
                            volume, connector, common,
                            host, remote_iscsi_ips, ready_ports,
                            target_portals, target_iqns, target_luns,
                            remote_client)

                        common._destroy_replication_client(remote_client)

                info = {'driver_volume_type': 'iscsi',
                        'data': {'target_portals': target_portals,
                                 'target_iqns': target_iqns,
                                 'target_luns': target_luns,
                                 'target_discovered': True
                                 }
                        }
            else:
                least_used_nsp = None

                # check if a VLUN already exists for this host
                existing_vlun = common.find_existing_vlun(volume, host)

                if existing_vlun:
                    # We override the nsp here on purpose to force the
                    # volume to be exported out the same IP as it already is.
                    # This happens during nova live-migration, we want to
                    # disable the picking of a different IP that we export
                    # the volume to, or nova complains.
                    least_used_nsp = common.build_nsp(existing_vlun['portPos'])

                if not least_used_nsp:
                    least_used_nsp = self._get_least_used_nsp_for_host(
                        common,
                        host['name'])

                vlun = None
                if existing_vlun is None:
                    # now that we have a host, create the VLUN
                    vlun = common.create_vlun(volume, host, least_used_nsp)
                else:
                    vlun = existing_vlun

                if least_used_nsp is None:
                    LOG.warning("Least busy iSCSI port not found, "
                                "using first iSCSI port in list.")
                    iscsi_ip = list(iscsi_ips)[0]
                else:
                    iscsi_ip = self._get_ip_using_nsp(least_used_nsp, common)

                iscsi_ip_port = iscsi_ips[iscsi_ip]['ip_port']
                iscsi_target_iqn = iscsi_ips[iscsi_ip]['iqn']
                info = {'driver_volume_type': 'iscsi',
                        'data': {'target_portal': "%s:%s" %
                                 (iscsi_ip, iscsi_ip_port),
                                 'target_iqn': iscsi_target_iqn,
                                 'target_lun': vlun['lun'],
                                 'target_discovered': True
                                 }
                        }

            if common._client_conf['hpe3par_iscsi_chap_enabled']:
                info['data']['auth_method'] = 'CHAP'
                info['data']['auth_username'] = username
                info['data']['auth_password'] = password

            encryption_key_id = volume.get('encryption_key_id', None)
            info['data']['encrypted'] = encryption_key_id is not None

            return info
        finally:
            self._logout(common)

    @utils.trace
    @coordination.synchronized('3par-{volume.id}')
    def terminate_connection(self, volume, connector, **kwargs):
        """Driver entry point to detach a volume from an instance."""
        array_id = self.get_volume_replication_driver_data(volume)
        common = self._login(array_id=array_id)
        try:
            is_force_detach = connector is None

            remote_client = None
            multipath = False
            if connector:
                multipath = connector.get('multipath')
            LOG.debug("multipath: %(multipath)s",
                      {'multipath': multipath})
            if multipath:
                if volume.get('replication_status') == 'enabled':
                    LOG.debug('This is a replication setup')

                    remote_target = common._replication_targets[0]
                    replication_mode = remote_target['replication_mode']
                    quorum_witness_ip = (
                        remote_target.get('quorum_witness_ip'))

                    if replication_mode == 1:
                        LOG.debug('replication_mode is sync')
                        if quorum_witness_ip:
                            LOG.debug('quorum_witness_ip is present')
                            LOG.debug('Peer Persistence has been configured')
                        else:
                            LOG.debug('Since quorum_witness_ip is absent, '
                                      'considering this as Active/Passive '
                                      'replication')
                    else:
                        LOG.debug('Active/Passive replication has been '
                                  'configured')

                    if replication_mode == 1 and quorum_witness_ip:
                        remote_client = (
                            common._create_replication_client(remote_target))

            if is_force_detach:
                common.terminate_connection(volume, None, None)
            else:
                hostname = common._safe_hostname(connector, self.configuration)
                common.terminate_connection(
                    volume,
                    hostname,
                    iqn=connector['initiator'],
                    remote_client=remote_client)
            self._clear_chap_3par(common, volume)
        finally:
            self._logout(common)

    def _clear_chap_3par(self, common, volume):
        """Clears CHAP credentials on a 3par volume.

        Ignore exceptions caused by the keys not being present on a volume.
        """
        vol_name = common._get_3par_vol_name(volume)

        try:
            common.client.removeVolumeMetaData(vol_name, CHAP_USER_KEY)
        except hpeexceptions.HTTPNotFound:
            pass
        except Exception:
            raise

        try:
            common.client.removeVolumeMetaData(vol_name, CHAP_PASS_KEY)
        except hpeexceptions.HTTPNotFound:
            pass
        except Exception:
            raise

    def _create_3par_iscsi_host(self, common, hostname, iscsi_iqn, domain,
                                persona_id, remote_client=None):
        """Create a 3PAR host.

        Create a 3PAR host, if there is already a host on the 3par using
        the same iqn but with a different hostname, return the hostname
        used by 3PAR.
        """
        # first search for an existing host
        host_found = None

        if remote_client:
            client_obj = remote_client
        else:
            client_obj = common.client

        hosts = client_obj.queryHost(iqns=iscsi_iqn)

        if hosts and hosts['members'] and 'name' in hosts['members'][0]:
            host_found = hosts['members'][0]['name']

        if host_found is not None:
            return host_found
        else:
            persona_id = int(persona_id)
            try:
                client_obj.createHost(hostname, iscsiNames=iscsi_iqn,
                                      optional={'domain': domain,
                                                'persona': persona_id})
            except hpeexceptions.HTTPConflict as path_conflict:
                msg = "Create iSCSI host caught HTTP conflict code: %s"
                with save_and_reraise_exception(reraise=False) as ctxt:
                    if path_conflict.get_code() is EXISTENT_PATH:
                        # Handle exception : EXISTENT_PATH - host WWN/iSCSI
                        # name already used by another host
                        hosts = client_obj.queryHost(iqns=iscsi_iqn)
                        if hosts and hosts['members'] and (
                                'name' in hosts['members'][0]):
                            hostname = hosts['members'][0]['name']
                        else:
                            # re-raise last caught exception
                            ctxt.reraise = True
                            LOG.exception(msg, path_conflict.get_code())
                    else:
                        # re-raise last caught exception
                        # for other HTTP conflict
                        ctxt.reraise = True
                        LOG.exception(msg, path_conflict.get_code())
            return hostname

    def _modify_3par_iscsi_host(self, common, hostname, iscsi_iqn):
        mod_request = {'pathOperation': common.client.HOST_EDIT_ADD,
                       'iSCSINames': [iscsi_iqn]}

        common.client.modifyHost(hostname, mod_request)

    def _set_3par_chaps(self, common, hostname, volume, username, password):
        """Sets a 3PAR host's CHAP credentials."""
        if not common._client_conf['hpe3par_iscsi_chap_enabled']:
            return

        mod_request = {'chapOperation': common.client.HOST_EDIT_ADD,
                       'chapOperationMode': common.client.CHAP_INITIATOR,
                       'chapName': username,
                       'chapSecret': password}
        common.client.modifyHost(hostname, mod_request)

    def _create_host(self, common, volume, connector,
                     remote_target=None, src_cpg=None, remote_client=None):
        """Creates or modifies existing 3PAR host."""
        # make sure we don't have the host already
        host = None
        domain = None
        username = None
        password = None
        hostname = common._safe_hostname(connector, self.configuration)

        if remote_target:
            cpg = common._get_cpg_from_cpg_map(
                remote_target['cpg_map'], src_cpg)
            cpg_obj = remote_client.getCPG(cpg)
            if 'domain' in cpg_obj:
                domain = cpg_obj['domain']
        else:
            cpg = common.get_cpg(volume, allowSnap=True)
            domain = common.get_domain(cpg)

        if not remote_target:
            # Get the CHAP secret if CHAP is enabled
            if common._client_conf['hpe3par_iscsi_chap_enabled']:
                vol_name = common._get_3par_vol_name(volume)
                username = common.client.getVolumeMetaData(
                    vol_name, CHAP_USER_KEY)['value']
                password = common.client.getVolumeMetaData(
                    vol_name, CHAP_PASS_KEY)['value']

        try:
            if remote_target:
                host = remote_client.getHost(hostname)
            else:
                host = common._get_3par_host(hostname)
                # Check whether host with iqn of initiator present on 3par
                hosts = common.client.queryHost(iqns=[connector['initiator']])
                host, hostname = (
                    common._get_prioritized_host_on_3par(
                        host, hosts, hostname))
        except hpeexceptions.HTTPNotFound:
            # get persona from the volume type extra specs
            persona_id = common.get_persona_type(volume)
            # host doesn't exist, we have to create it
            hostname = self._create_3par_iscsi_host(common,
                                                    hostname,
                                                    [connector['initiator']],
                                                    domain,
                                                    persona_id,
                                                    remote_client)
        else:
            if not remote_target:
                if 'iSCSIPaths' not in host or len(host['iSCSIPaths']) < 1:
                    self._modify_3par_iscsi_host(
                        common, hostname,
                        connector['initiator'])
                elif (not host['initiatorChapEnabled'] and
                        common._client_conf['hpe3par_iscsi_chap_enabled']):
                    LOG.warning("Host exists without CHAP credentials set and "
                                "has iSCSI attachments but CHAP is enabled. "
                                "Updating host with new CHAP credentials.")

        if remote_target:
            host = remote_client.getHost(hostname)
        else:
            # set/update the chap details for the host
            self._set_3par_chaps(common, hostname, volume, username, password)
            host = common._get_3par_host(hostname)
        return host, username, password, cpg

    def _do_export(self, common, volume, connector):
        """Gets the associated account, generates CHAP info and updates."""
        model_update = {}

        if not common._client_conf['hpe3par_iscsi_chap_enabled']:
            model_update['provider_auth'] = None
            return model_update

        # CHAP username will be the hostname
        chap_username = connector['host']

        chap_password = None
        try:
            # Get all active VLUNs for the host
            vluns = common.client.getHostVLUNs(chap_username)

            # Host has active VLUNs... is CHAP enabled on host?
            host_info = common.client.getHost(chap_username)

            if not host_info['initiatorChapEnabled']:
                LOG.warning("Host has no CHAP key, but CHAP is enabled.")

        except hpeexceptions.HTTPNotFound:
            chap_password = volume_utils.generate_password(16)
            LOG.warning("No host or VLUNs exist. Generating new "
                        "CHAP key.")
        else:
            # Get a list of all iSCSI VLUNs and see if there is already a CHAP
            # key assigned to one of them.  Use that CHAP key if present,
            # otherwise create a new one.  Skip any VLUNs that are missing
            # CHAP credentials in metadata.
            chap_exists = False
            active_vluns = 0

            for vlun in vluns:
                if not vlun['active']:
                    continue

                active_vluns += 1

                # iSCSI connections start with 'iqn'.
                if ('remoteName' in vlun and
                        re.match('iqn.*', vlun['remoteName'])):
                    try:
                        chap_password = common.client.getVolumeMetaData(
                            vlun['volumeName'], CHAP_PASS_KEY)['value']
                        chap_exists = True
                        break
                    except hpeexceptions.HTTPNotFound:
                        LOG.debug("The VLUN %s is missing CHAP credentials "
                                  "but CHAP is enabled. Skipping.",
                                  vlun['remoteName'])
                else:
                    LOG.warning("Non-iSCSI VLUN detected.")

            if not chap_exists:
                chap_password = volume_utils.generate_password(16)
                LOG.warning("No VLUN contained CHAP credentials. "
                            "Generating new CHAP key.")

        # Add CHAP credentials to the volume metadata
        vol_name = common._get_3par_vol_name(volume)
        common.client.setVolumeMetaData(
            vol_name, CHAP_USER_KEY, chap_username)
        common.client.setVolumeMetaData(
            vol_name, CHAP_PASS_KEY, chap_password)

        model_update['provider_auth'] = ('CHAP %s %s' %
                                         (chap_username, chap_password))

        return model_update

    @utils.trace
    def create_export(self, context, volume, connector):
        common = self._login()
        try:
            return self._do_export(common, volume, connector)
        finally:
            self._logout(common)

    @utils.trace
    def ensure_export(self, context, volume):
        """Ensure the volume still exists on the 3PAR.

        Also retrieves CHAP credentials, if present on the volume
        """
        common = self._login()
        try:
            vol_name = common._get_3par_vol_name(volume)
            common.client.getVolume(vol_name)
        except hpeexceptions.HTTPNotFound:
            LOG.error("Volume %s doesn't exist on array.", vol_name)
        else:
            metadata = common.client.getAllVolumeMetaData(vol_name)

            username = None
            password = None
            model_update = {}
            model_update['provider_auth'] = None

            for member in metadata['members']:
                if member['key'] == CHAP_USER_KEY:
                    username = member['value']
                elif member['key'] == CHAP_PASS_KEY:
                    password = member['value']

            if username and password:
                model_update['provider_auth'] = ('CHAP %s %s' %
                                                 (username, password))

            return model_update
        finally:
            self._logout(common)

    def _get_least_used_nsp_for_host(self, common, hostname):
        """Get the least used NSP for the current host.

        Steps to determine which NSP to use.
            * If only one iSCSI NSP, return it
            * If there is already an active vlun to this host, return its NSP
            * Return NSP with fewest active vluns
        """

        iscsi_nsps = self._get_iscsi_nsps(common)
        # If there's only one path, use it
        if len(iscsi_nsps) == 1:
            return iscsi_nsps[0]

        # Try to reuse an existing iscsi path to the host
        vluns = common.client.getVLUNs()
        for vlun in vluns['members']:
            if vlun['active']:
                if vlun['hostname'] == hostname:
                    temp_nsp = common.build_nsp(vlun['portPos'])
                    if temp_nsp in iscsi_nsps:
                        # this host already has an iscsi path, so use it
                        return temp_nsp

        # Calculate the least used iscsi nsp
        least_used_nsp = self._get_least_used_nsp(common,
                                                  vluns['members'],
                                                  self._get_iscsi_nsps(common))
        return least_used_nsp

    def _get_iscsi_nsps(self, common):
        """Return the list of candidate nsps."""
        nsps = []
        iscsi_ips = self.iscsi_ips[common._client_conf['hpe3par_api_url']]
        for value in iscsi_ips.values():
            nsps.append(value['nsp'])
        return nsps

    def _get_ip_using_nsp(self, nsp, common):
        """Return IP associated with given nsp."""
        iscsi_ips = self.iscsi_ips[common._client_conf['hpe3par_api_url']]
        for (key, value) in iscsi_ips.items():
            if value['nsp'] == nsp:
                return key

    def _get_least_used_nsp(self, common, vluns, nspss):
        """Return the nsp that has the fewest active vluns."""
        # return only the nsp (node:server:port)
        # count the number of nsps
        nsp_counts = {}
        for nsp in nspss:
            # initialize counts to zero
            nsp_counts[nsp] = 0

        current_least_used_nsp = None

        for vlun in vluns:
            if vlun['active']:
                nsp = common.build_nsp(vlun['portPos'])
                if nsp in nsp_counts:
                    nsp_counts[nsp] = nsp_counts[nsp] + 1

        # identify key (nsp) of least used nsp
        current_smallest_count = sys.maxsize
        for (nsp, count) in nsp_counts.items():
            if count < current_smallest_count:
                current_least_used_nsp = nsp
                current_smallest_count = count
        return current_least_used_nsp



# Copyright 2014 IBM Corp.
# Copyright 2015 Clinton Knight
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import six

from cinder.api.openstack import api_version_request
from cinder import exception
from cinder.tests.unit import test


@ddt.ddt
class APIVersionRequestTests(test.TestCase):

    def test_init(self):

        result = api_version_request.APIVersionRequest()

        self.assertIsNone(result._ver_major)
        self.assertIsNone(result._ver_minor)

    def test_min_version(self):

        self.assertEqual(
            api_version_request.APIVersionRequest(
                api_version_request._MIN_API_VERSION),
            api_version_request.min_api_version())

    def test_max_api_version(self):

        self.assertEqual(
            api_version_request.APIVersionRequest(
                api_version_request._MAX_API_VERSION),
            api_version_request.max_api_version())

    @ddt.data(
        ('1.1', 1, 1),
        ('2.10', 2, 10),
        ('5.234', 5, 234),
        ('12.5', 12, 5),
        ('2.0', 2, 0),
        ('2.200', 2, 200)
    )
    @ddt.unpack
    def test_valid_version_strings(self, version_string, major, minor):

        request = api_version_request.APIVersionRequest(version_string)

        self.assertEqual(major, request._ver_major)
        self.assertEqual(minor, request._ver_minor)

    def test_null_version(self):
        v = api_version_request.APIVersionRequest()
        self.assertFalse(bool(v))

    def test_not_null_version(self):
        v = api_version_request.APIVersionRequest('1.1')
        self.assertTrue(bool(v))

    @ddt.data('2', '200', '2.1.4', '200.23.66.3', '5 .3', '5. 3',
              '5.03', '02.1', '2.001', '', ' 2.1', '2.1 ')
    def test_invalid_version_strings(self, version_string):

        self.assertRaises(exception.InvalidAPIVersionString,
                          api_version_request.APIVersionRequest,
                          version_string)

    def test_cmpkey(self):
        request = api_version_request.APIVersionRequest('1.2')
        self.assertEqual((1, 2), request._cmpkey())

    def test_version_comparisons(self):
        v1 = api_version_request.APIVersionRequest('2.0')
        v2 = api_version_request.APIVersionRequest('2.5')
        v3 = api_version_request.APIVersionRequest('5.23')
        v4 = api_version_request.APIVersionRequest('2.0')
        v_null = api_version_request.APIVersionRequest()

        self.assertLess(v1, v2)
        self.assertLessEqual(v1, v2)
        self.assertGreater(v3, v2)
        self.assertGreaterEqual(v3, v2)
        self.assertNotEqual(v1, v2)
        self.assertEqual(v1, v4)
        self.assertNotEqual(v1, v_null)
        self.assertEqual(v_null, v_null)
        self.assertNotEqual('2.0', v1)

    def test_version_matches(self):
        v1 = api_version_request.APIVersionRequest('2.0')
        v2 = api_version_request.APIVersionRequest('2.5')
        v3 = api_version_request.APIVersionRequest('2.45')
        v4 = api_version_request.APIVersionRequest('3.3')
        v5 = api_version_request.APIVersionRequest('3.23')
        v6 = api_version_request.APIVersionRequest('2.0')
        v7 = api_version_request.APIVersionRequest('3.3')
        v8 = api_version_request.APIVersionRequest('4.0')
        v_null = api_version_request.APIVersionRequest()

        self.assertTrue(v2.matches(v1, v3))
        self.assertTrue(v2.matches(v1, v_null))
        self.assertTrue(v1.matches(v6, v2))
        self.assertTrue(v4.matches(v2, v7))
        self.assertTrue(v4.matches(v_null, v7))
        self.assertTrue(v4.matches(v_null, v8))
        self.assertFalse(v1.matches(v2, v3))
        self.assertFalse(v5.matches(v2, v4))
        self.assertFalse(v2.matches(v3, v1))
        self.assertTrue(v1.matches(v_null, v_null))

        self.assertRaises(ValueError, v_null.matches, v1, v3)

    def test_matches_versioned_method(self):

        request = api_version_request.APIVersionRequest('2.0')

        self.assertRaises(exception.InvalidParameterValue,
                          request.matches_versioned_method,
                          'fake_method')

    def test_get_string(self):
        v1_string = '3.23'
        v1 = api_version_request.APIVersionRequest(v1_string)
        self.assertEqual(v1_string, v1.get_string())

        self.assertRaises(ValueError,
                          api_version_request.APIVersionRequest().get_string)

    @ddt.data(('1', '0'), ('1', '1'))
    @ddt.unpack
    def test_str(self, major, minor):
        request_input = '%s.%s' % (major, minor)
        request = api_version_request.APIVersionRequest(request_input)
        request_string = six.text_type(request)

        self.assertEqual('API Version Request '
                         'Major: %s, Minor: %s' % (major, minor),
                         request_string)



# Copyright 2020 Datera
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
from unittest import mock
import uuid

from cinder import context
from cinder import exception
from cinder.tests.unit import test
from cinder import version
from cinder.volume import configuration as conf
from cinder.volume import volume_types

sys.modules['dfs_sdk'] = mock.MagicMock()

from cinder.volume.drivers.datera import datera_iscsi as datera  # noqa

datera.datc.DEFAULT_SI_SLEEP = 0
datera.datc.DEFAULT_SNAP_SLEEP = 0
OS_PREFIX = datera.datc.OS_PREFIX
UNMANAGE_PREFIX = datera.datc.UNMANAGE_PREFIX
DateraAPIException = datera.datc.DateraAPIException


class DateraVolumeTestCasev22(test.TestCase):

    def setUp(self):
        self.cfg = mock.Mock(spec=conf.Configuration)
        self.cfg.san_ip = '127.0.0.1'
        self.cfg.datera_api_port = '7717'
        self.cfg.san_is_local = True
        self.cfg.datera_num_replicas = 1
        self.cfg.datera_503_timeout = 0.01
        self.cfg.datera_503_interval = 0.001
        self.cfg.datera_debug = False
        self.cfg.san_login = 'user'
        self.cfg.san_password = 'pass'
        self.cfg.datera_tenant_id = '/root/test-tenant'
        self.cfg.driver_client_cert = None
        self.cfg.driver_client_cert_key = None
        self.cfg.datera_disable_profiler = False
        self.cfg.datera_ldap_server = ""
        self.cfg.datera_volume_type_defaults = {}
        self.cfg.datera_disable_template_override = False
        self.cfg.datera_disable_extended_metadata = False
        self.cfg.datera_enable_image_cache = False
        self.cfg.datera_image_cache_volume_type_id = ""
        self.cfg.filter_function = lambda: None
        self.cfg.goodness_function = lambda: None
        self.cfg.use_chap_auth = False
        self.cfg.chap_username = ""
        self.cfg.chap_password = ""

        super(DateraVolumeTestCasev22, self).setUp()
        mock_exec = mock.Mock()
        mock_exec.return_value = ('', '')

        self.driver = datera.DateraDriver(execute=mock_exec,
                                          configuration=self.cfg)
        self.driver.api = mock.MagicMock()
        self.driver.apiv = "2.2"

        self.driver.set_initialized()
        # No-op config getter
        self.driver.configuration.get = lambda *args, **kwargs: {}
        # self.addCleanup(self.api_patcher.stop)
        self.driver.datera_version = "3.3.3"

    def test_volume_create_success(self):
        testvol = _stub_volume()
        self.assertIsNone(self.driver.create_volume(testvol))

    def test_volume_create_fails(self):
        testvol = _stub_volume()
        self.driver.api.app_instances.create.side_effect = DateraAPIException
        self.assertRaises(DateraAPIException,
                          self.driver.create_volume,
                          testvol)

    @mock.patch.object(volume_types, 'get_volume_type')
    def test_create_volume_with_extra_specs(self, mock_get_type):
        mock_get_type.return_value = {
            'name': u'The Best',
            'qos_specs_id': None,
            'deleted': False,
            'created_at': '2015-08-14 04:18:11',
            'updated_at': None,
            'extra_specs': {
                u'volume_backend_name': u'datera',
                u'qos:max_iops_read': u'2000',
                u'qos:max_iops_write': u'4000',
                u'qos:max_iops_total': u'4000'
            },
            'is_public': True,
            'deleted_at': None,
            'id': u'dffb4a83-b8fb-4c19-9f8c-713bb75db3b1',
            'description': None
        }

        mock_volume = _stub_volume(
            volume_type_id='dffb4a83-b8fb-4c19-9f8c-713bb75db3b1'
        )

        self.assertIsNone(self.driver.create_volume(mock_volume))
        self.assertTrue(mock_get_type.called)

    def test_create_cloned_volume_success(self):
        testvol = _stub_volume()
        ref = _stub_volume(id=str(uuid.uuid4()))
        self.assertIsNone(self.driver.create_cloned_volume(testvol, ref))

    def test_create_cloned_volume_success_larger(self):
        newsize = 2
        testvol = _stub_volume(size=newsize)
        ref = _stub_volume(id=str(uuid.uuid4()))
        mock_extend = mock.MagicMock()
        self.driver._extend_volume_2_2 = mock_extend
        self.driver._extend_volume_2_1 = mock_extend
        self.driver.create_cloned_volume(testvol, ref)
        mock_extend.assert_called_once_with(testvol, newsize)

    def test_create_cloned_volume_fails(self):
        testvol = _stub_volume()
        ref = _stub_volume(id=str(uuid.uuid4()))
        self.driver.api.app_instances.create.side_effect = DateraAPIException
        self.assertRaises(DateraAPIException,
                          self.driver.create_cloned_volume,
                          testvol,
                          ref)

    def test_delete_volume_success(self):
        testvol = _stub_volume()
        self.driver.api.app_instances.delete.return_value = {}
        self.assertIsNone(self.driver.delete_volume(testvol))

    def test_delete_volume_not_found(self):
        testvol = _stub_volume()
        self.driver.api.app_instances.list.side_effect = exception.NotFound
        self.assertIsNone(self.driver.delete_volume(testvol))

    def test_delete_volume_fails(self):
        testvol = _stub_volume()
        self.driver.api.app_instances.list.side_effect = DateraAPIException
        self.assertRaises(DateraAPIException,
                          self.driver.delete_volume, testvol)

    def test_ensure_export_success(self):
        testvol = _stub_volume()
        ctxt = context.get_admin_context()
        self.assertIsNone(self.driver.ensure_export(ctxt, testvol, None))

    def test_ensure_export_fails(self):
        # This can't fail because it's a no-op
        testvol = _stub_volume()
        ctxt = context.get_admin_context()
        self.assertIsNone(self.driver.ensure_export(ctxt, testvol, None))

    def test_create_export_target_does_not_exist_success(self):
        testvol = _stub_volume()
        aimock = mock.MagicMock()
        simock = mock.MagicMock()
        simock.reload.return_value = simock
        aimock.storage_instances.list.return_value = [simock]
        simock.op_state = "available"
        self.driver.cvol_to_ai = mock.Mock()
        self.driver.cvol_to_ai.return_value = aimock
        self.assertIsNone(self.driver.create_export(None, testvol, None))

    def test_create_export_fails(self):
        testvol = _stub_volume()
        aimock = mock.MagicMock()
        simock = mock.MagicMock()
        simock.reload.return_value = simock
        aimock.storage_instances.list.side_effect = DateraAPIException
        simock.op_state = "available"
        self.driver.cvol_to_ai = mock.Mock()
        self.driver.cvol_to_ai.return_value = aimock
        self.assertRaises(DateraAPIException,
                          self.driver.create_export,
                          None,
                          testvol,
                          None)

    def test_initialize_connection_success(self):
        testvol = _stub_volume()
        aimock = mock.MagicMock()
        simock = mock.MagicMock()
        simock.access = {"ips": ["test-ip"], "iqn": "test-iqn"}
        simock.reload.return_value = simock
        aimock.storage_instances.list.return_value = [simock]
        self.driver.cvol_to_ai = mock.Mock()
        self.driver.cvol_to_ai.return_value = aimock
        self.assertEqual(self.driver.initialize_connection(testvol, {}),
                         {'data': {'discard': False,
                                   'target_discovered': False,
                                   'target_iqn': 'test-iqn',
                                   'target_lun': 0,
                                   'target_portal': 'test-ip:3260',
                                   'volume_id': testvol['id']},
                          'driver_volume_type': 'iscsi'})

    def test_initialize_connection_fails(self):
        testvol = _stub_volume()
        aimock = mock.MagicMock()
        simock = mock.MagicMock()
        simock.access = {"ips": ["test-ip"], "iqn": "test-iqn"}
        simock.reload.return_value = simock
        aimock.storage_instances.list.side_effect = DateraAPIException
        self.driver.cvol_to_ai = mock.Mock()
        self.driver.cvol_to_ai.return_value = aimock
        self.assertRaises(DateraAPIException,
                          self.driver.initialize_connection,
                          testvol,
                          {})

    def test_detach_volume_success(self):
        testvol = _stub_volume()
        self.driver.cvol_to_ai = mock.MagicMock()
        aimock = mock.MagicMock()
        aimock.set.return_value = {}
        self.driver.cvol_to_ai.return_value = aimock
        ctxt = context.get_admin_context()
        self.assertIsNone(self.driver.detach_volume(ctxt, testvol))

    def test_detach_volume_fails(self):
        testvol = _stub_volume()
        self.driver.cvol_to_ai = mock.MagicMock()
        aimock = mock.MagicMock()
        aimock.set.side_effect = DateraAPIException
        self.driver.cvol_to_ai.return_value = aimock
        ctxt = context.get_admin_context()
        self.assertRaises(DateraAPIException,
                          self.driver.detach_volume,
                          ctxt, testvol)

    def test_detach_volume_not_found(self):
        testvol = _stub_volume()
        self.driver.cvol_to_ai = mock.MagicMock()
        aimock = mock.MagicMock()
        aimock.set.side_effect = exception.NotFound
        self.driver.cvol_to_ai.return_value = aimock
        ctxt = context.get_admin_context()
        self.assertIsNone(self.driver.detach_volume(ctxt, testvol))

    def test_create_snapshot_success(self):
        testsnap = _stub_snapshot(volume_id=str(uuid.uuid4()))
        volmock = mock.MagicMock()
        snapmock = mock.MagicMock()
        snapmock.reload.return_value = snapmock
        snapmock.uuid = testsnap['id']
        snapmock.op_state = "available"
        volmock.snapshots.create.return_value = snapmock
        self.driver.cvol_to_dvol = mock.MagicMock()
        self.driver.cvol_to_dvol.return_value = volmock
        self.assertIsNone(self.driver.create_snapshot(testsnap))

    def test_create_snapshot_fails(self):
        testsnap = _stub_snapshot(volume_id=str(uuid.uuid4()))
        self.driver.api.app_instances.list.side_effect = DateraAPIException
        self.assertRaises(DateraAPIException,
                          self.driver.create_snapshot,
                          testsnap)

    def test_delete_snapshot_success(self):
        testsnap = _stub_snapshot(volume_id=str(uuid.uuid4()))
        self.assertIsNone(self.driver.delete_snapshot(testsnap))

    def test_delete_snapshot_not_found(self):
        testsnap = _stub_snapshot(volume_id=str(uuid.uuid4()))
        self.driver.cvol_to_dvol = mock.MagicMock()
        aimock = mock.MagicMock()
        aimock.snapshots.list.side_effect = exception.NotFound
        self.driver.cvol_to_dvol.return_value = aimock
        self.assertIsNone(self.driver.delete_snapshot(testsnap))

    def test_delete_snapshot_fails(self):
        testsnap = _stub_snapshot(volume_id=str(uuid.uuid4()))
        self.driver.cvol_to_dvol = mock.MagicMock()
        aimock = mock.MagicMock()
        aimock.snapshots.list.side_effect = DateraAPIException
        self.driver.cvol_to_dvol.return_value = aimock
        self.assertRaises(DateraAPIException,
                          self.driver.delete_snapshot,
                          testsnap)

    def test_create_volume_from_snapshot_success(self):
        testsnap = _stub_snapshot(volume_id=str(uuid.uuid4()))
        testvol = _stub_volume()
        volmock = mock.MagicMock()
        snapmock = mock.MagicMock()
        snapmock.reload.return_value = snapmock
        snapmock.uuid = testsnap['id']
        snapmock.op_state = "available"
        self.driver.cvol_to_dvol = mock.MagicMock()
        self.driver.cvol_to_dvol.return_value = volmock
        volmock.snapshots.list.return_value = [snapmock]
        self.assertIsNone(self.driver.create_volume_from_snapshot(
            testvol, testsnap))

    def test_create_volume_from_snapshot_fails(self):
        testsnap = _stub_snapshot(volume_id=str(uuid.uuid4()))
        testvol = _stub_volume()
        self.driver.cvol_to_dvol = mock.MagicMock()
        aimock = mock.MagicMock()
        aimock.snapshots.list.side_effect = DateraAPIException
        self.driver.cvol_to_dvol.return_value = aimock
        self.assertRaises(DateraAPIException,
                          self.driver.create_volume_from_snapshot,
                          testvol,
                          testsnap)

    def test_extend_volume_success(self):
        newsize = 2
        testvol = _stub_volume()
        mockvol = mock.MagicMock()
        mockvol.size = newsize
        self.driver.cvol_to_dvol = mock.MagicMock()
        self.driver.cvol_to_dvol.return_value = mockvol
        self.driver._offline_flip_2_2 = mock.MagicMock()
        self.driver._offline_flip_2_1 = mock.MagicMock()
        self.assertIsNone(self.driver.extend_volume(testvol, newsize))

    def test_extend_volume_fails(self):
        newsize = 2
        testvol = _stub_volume()
        mockvol = mock.MagicMock()
        mockvol.size = newsize
        mockvol.set.side_effect = DateraAPIException
        self.driver.cvol_to_dvol = mock.MagicMock()
        self.driver.cvol_to_dvol.return_value = mockvol
        self.driver._offline_flip_2_2 = mock.MagicMock()
        self.driver._offline_flip_2_1 = mock.MagicMock()
        self.assertRaises(DateraAPIException,
                          self.driver.extend_volume,
                          testvol,
                          newsize)

    def test_manage_existing(self):
        existing_ref = {'source-name': "A:B:C:D"}
        testvol = _stub_volume()
        self.driver.cvol_to_ai = mock.MagicMock()
        self.assertIsNone(self.driver.manage_existing(testvol, existing_ref))

    def test_manage_existing_wrong_ref(self):
        existing_ref = {'source-name': "ABCD"}
        testvol = _stub_volume()
        self.driver.cvol_to_ai = mock.MagicMock()
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing,
                          testvol,
                          existing_ref)

    def test_manage_existing_get_size(self):
        existing_ref = {'source-name': "A:B:C:D"}
        testvol = _stub_volume()
        volmock = mock.MagicMock()
        volmock.size = testvol['size']
        self.driver.cvol_to_dvol = mock.MagicMock()
        self.driver.cvol_to_dvol.return_value = volmock
        self.assertEqual(self.driver.manage_existing_get_size(
            testvol, existing_ref), testvol['size'])

    def test_manage_existing_get_size_wrong_ref(self):
        existing_ref = {'source-name': "ABCD"}
        testvol = _stub_volume()
        self.driver.cvol_to_ai = mock.MagicMock()
        self.assertRaises(exception.ManageExistingInvalidReference,
                          self.driver.manage_existing_get_size,
                          testvol,
                          existing_ref)

    def test_get_manageable_volumes(self):
        testvol = _stub_volume()
        v1 = {'reference': {'source-name': 'some-ai:storage-1:volume-1'},
              'size': 1,
              'safe_to_manage': True,
              'reason_not_safe': '',
              'cinder_id': None,
              'extra_info': {'snapshots': '[]'}}
        v2 = {'reference': {'source-name': 'some-other-ai:storage-1:volume-1'},
              'size': 2,
              'safe_to_manage': True,
              'reason_not_safe': '',
              'cinder_id': None,
              'extra_info': {'snapshots': '[]'}}

        mock1 = mock.MagicMock()
        mock1.__getitem__.side_effect = ['some-ai']
        mock1.name = 'some-ai'
        mocksi1 = mock.MagicMock()
        mocksi1.name = "storage-1"
        mocksi1.__getitem__.side_effect = [[mock.MagicMock()]]
        mock1.storage_instances.list.return_value = [mocksi1]
        mockvol1 = mock.MagicMock()
        mockvol1.name = "volume-1"
        mockvol1.size = v1['size']
        mocksi1.volumes.list.return_value = [mockvol1]

        mock2 = mock.MagicMock()
        mock2.__getitem__.side_effect = ['some-other-ai']
        mock2.name = 'some-other-ai'
        mocksi2 = mock.MagicMock()
        mocksi2.name = "storage-1"
        mocksi2.__getitem__.side_effect = [[mock.MagicMock()]]
        mock2.storage_instances.list.return_value = [mocksi2]
        mockvol2 = mock.MagicMock()
        mockvol2.name = "volume-1"
        mockvol2.size = v2['size']
        mocksi2.volumes.list.return_value = [mockvol2]

        listmock = mock.MagicMock()
        listmock.return_value = [mock1, mock2]
        self.driver.api.app_instances.list = listmock

        marker = mock.MagicMock()
        limit = mock.MagicMock()
        offset = mock.MagicMock()
        sort_keys = mock.MagicMock()
        sort_dirs = mock.MagicMock()
        if (version.version_string() >= '15.0.0'):
            with mock.patch(
                    'cinder.volume.volume_utils.paginate_entries_list') \
                    as mpage:
                self.driver.get_manageable_volumes(
                    [testvol], marker, limit, offset, sort_keys, sort_dirs)
                mpage.assert_called_once_with(
                    [v1, v2], marker, limit, offset, sort_keys, sort_dirs)
        else:
            with mock.patch(
                    'cinder.volume.utils.paginate_entries_list') as mpage:
                self.driver.get_manageable_volumes(
                    [testvol], marker, limit, offset, sort_keys, sort_dirs)
                mpage.assert_called_once_with(
                    [v1, v2], marker, limit, offset, sort_keys, sort_dirs)

    def test_unmanage(self):
        testvol = _stub_volume()
        self.assertIsNone(self.driver.unmanage(testvol))


class DateraVolumeTestCasev21(DateraVolumeTestCasev22):

    def setUp(self):
        super(DateraVolumeTestCasev21, self).setUp()
        self.driver.api = mock.MagicMock()
        self.driver.apiv = '2.1'


def _stub_volume(*args, **kwargs):
    uuid = 'c20aba21-6ef6-446b-b374-45733b4883ba'
    name = 'volume-00000001'
    size = 1
    volume = {}
    volume['id'] = kwargs.get('id', uuid)
    volume['project_id'] = "test-project"
    volume['display_name'] = kwargs.get('display_name', name)
    volume['size'] = kwargs.get('size', size)
    volume['provider_location'] = kwargs.get('provider_location', None)
    volume['volume_type_id'] = kwargs.get('volume_type_id', None)
    return volume


def _stub_snapshot(*args, **kwargs):
    uuid = '0bb34f0c-fea4-48e0-bf96-591120ac7e3c'
    name = 'snapshot-00000001'
    volume_size = 1
    snap = {}
    snap['id'] = kwargs.get('id', uuid)
    snap['project_id'] = "test-project"
    snap['display_name'] = kwargs.get('display_name', name)
    snap['volume_id'] = kwargs.get('volume_id', None)
    snap['volume_size'] = kwargs.get('volume_size', volume_size)
    return snap



# Copyright (c) 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils
from oslo_versionedobjects import fields

from cinder.db.sqlalchemy import models
from cinder import objects


def cluster_basic_fields():
    """Return basic fields for a cluster."""
    return {
        'id': 1,
        'created_at': timeutils.utcnow(with_timezone=False),
        'deleted': False,
        'name': 'cluster_name',
        'binary': 'cinder-volume',
        'race_preventer': 0,
    }


def fake_cluster_orm(**updates):
    """Create a fake ORM cluster instance."""
    db_cluster = fake_db_cluster(**updates)
    del db_cluster['services']
    cluster = models.Cluster(**db_cluster)
    return cluster


def fake_db_cluster(**updates):
    """Helper method for fake_cluster_orm.

    Creates a complete dictionary filling missing fields based on the Cluster
    field definition (defaults and nullable).
    """
    db_cluster = cluster_basic_fields()

    for name, field in objects.Cluster.fields.items():
        if name in db_cluster:
            continue
        if field.default != fields.UnspecifiedDefault:
            db_cluster[name] = field.default
        elif field.nullable:
            db_cluster[name] = None
        else:
            raise Exception('fake_db_cluster needs help with %s.' % name)

    if updates:
        db_cluster.update(updates)

    return db_cluster


def fake_cluster_ovo(context, **updates):
    """Create a fake Cluster versioned object."""
    return objects.Cluster._from_db_object(context, objects.Cluster(),
                                           fake_cluster_orm(**updates))
